
from epyk.core.js import JsUtils
from epyk.core.js.primitives import JsObjects
from epyk.core.js.packages import DataAttrs
from epyk.core.js.packages import JsPackage

from epyk.core.css import Colors


EASING_OPTIONS = ['linear', 'easeInQuad', 'easeOutQuad', 'easeInOutQuad', 'easeInCubic', 'easeOutCubic',
                  'easeInOutCubic', 'easeInQuart', 'easeOutQuart', 'easeInOutQuart', 'easeInQuint', 'easeOutQuint',
                  'easeInOutQuint', 'easeInSine', 'easeOutSine', 'easeInOutSine', 'easeInExpo', 'easeOutExpo',
                  'easeInOutExpo', 'easeInCirc', 'easeOutCirc', 'easeInOutCirc', 'easeInElastic', 'easeOutElastic',
                  'easeInOutElastic', 'easeInBack', 'easeOutBack', 'easeInOutBack', 'easeInBounce', 'easeOutBounce',
                  'easeInOutBounce']


class Config(DataAttrs):

  @property
  def duration(self):
    """
    Description:
    ------------
    Time for the animation of the redraw in milliseconds

    Related Pages:

			https://www.chartjs.org/docs/latest/developers/api.html
    """
    return self._attrs["duration"]

  @duration.setter
  def duration(self, time):
    self._attrs["duration"] = time

  @property
  def lazy(self):
    """
    Description:
    ------------
    If true, the animation can be interrupted by other animations

    Related Pages:

			https://www.chartjs.org/docs/latest/developers/api.html
    """
    return self._attrs["lazy"]

  @lazy.setter
  def lazy(self, flag):
    self._attrs["lazy"] = flag

  @property
  def easing(self):
    """
    Description:
    ------------
    Set the easing option

    Related Pages:

			https://www.chartjs.org/docs/latest/configuration/animations.html
    """
    return self._attrs["easing"]

  @easing.setter
  def easing(self, value):
    if value not in EASING_OPTIONS:
      raise Exception("%s is not allowed" % value)

    self._attrs["easing"] = value


class ChartJs(JsPackage):
  lib_alias = {'js': 'Chart.js'}

  class __internal(object):
    jqId, htmlCode, jsImports, cssImport = 'chart', '', set([]), set([])

  def __init__(self, htmlCode=None, config=None, src=None, varName=None, selector=None, setVar=False):
    self.src = src if src is not None else self.__internal()
    if selector is None:
      self._selector = 'new Chart(%s.getContext("2d"), %s)' % (htmlCode, config.toStr())
    else:
      self._selector = selector
    self.varName, self.setVar = varName or self._selector, setVar
    self.src.jsImports.add(self.lib_alias['js'])
    self._js = []

  def getElementsAtEvent(self, jsEvent):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param jsEvent:
    """
    return JsObjects.JsArray.JsArray("%s.getElementsAtEvent(%s)" % (self.varName, jsEvent), isPyData=False)

  def add(self, point, values):
    """
    Description:
    -----------
    Add point to an exiting chart on existing series

    Note: Do not forget to trigger an update on the chart once all your transformations are done

    Related Pages:

      https://www.chartjs.org/docs/latest/developers/updates.html

    Attributes:
    ----------
    :param point: Object. The point to add on the x-axis
    :param values: dictionary. The value per series name
    """
    point = JsUtils.jsConvertData(point, None)
    values = JsUtils.jsConvertData(values, None)
    return JsObjects.JsVoid('''%(varName)s.data.labels.push(%(point)s);
      %(varName)s.data.datasets.forEach(function(dataset){
        dataset.data.push(%(values)s[dataset.label])})''' % {'varName': self.varName, 'point': point, 'values': values})

  def remove(self, point=None, seriesNames=None):
    """
    Description:
    -----------
    Remove point to existing series

    Note: Do not forget to trigger an update on the chart once all your transformations are done

    Related Pages:

      https://www.chartjs.org/docs/latest/developers/updates.html

    Attributes:
    ----------
    :param point: Object. Optional. The point to be removed on the series. If none the last one will be removed
    :param seriesNames: List. Optional. The series name
    """
    point = JsUtils.jsConvertData(point, None)
    if seriesNames is None:
      if point is None:
        return JsObjects.JsVoid('''
          %(varName)s.data.labels.pop();
          %(varName)s.data.datasets.forEach(function(dataset){ dataset.data.pop() })''' % {'varName': self.varName})

      return JsObjects.JsVoid('''
        var index = %(varName)s.data.labels.indexOf(%(point)s);
        if(index >= 0){
          %(varName)s.data.labels.splice(index, 1);
          %(varName)s.data.datasets.forEach(function(dataset){ dataset.data.splice(index, 1) })
        } ''' % {'varName': self.varName, 'point': point})

    seriesNames = JsUtils.jsConvertData(seriesNames, None)
    return JsObjects.JsVoid('''
        var index = %(varName)s.data.labels.indexOf(%(point)s);
        if(index >= 0){
          %(varName)s.data.datasets.forEach(function(dataset){
            if (%(seriesNames)s.includes(dataset.label)){
              dataset.data[index] = null}})} ''' % {'varName': self.varName, 'point': point, 'seriesNames': seriesNames})

  def load(self, name, points, options=None):
    """
    Description:
    -----------
    Loads new series on an existing chart.
    existing x axis will not be changed and they will be used to add the points

    Note: Do not forget to trigger an update on the chart once all your transformations are done

    Related Pages:

      https://www.chartjs.org/docs/latest/developers/updates.html

    Attributes:
    ----------
    :param name: String. The series name
    :param points: List of dictionaries. The list of points ({x: , y: }) to be added to the chart
    """
    name = JsUtils.jsConvertData(name, None)
    points = JsUtils.jsConvertData(points, None)
    return JsObjects.JsVoid('''
      var values = []; var index= -1;
      %(varName)s.data.datasets.forEach(function(d, i){ if(d.label == %(name)s){ index = i}});
      if (index == -1){
        %(varName)s.data.labels.forEach(function(v){ values.push(%(points)s.v) });
        %(varName)s.data.datasets.push({label: %(name)s, data: %(points)s})
      }''' % {'varName': self.varName, 'name': name, 'points': points})

  def unload(self, names=None):
    """
    Description:
    -----------
    Remove series from the chart

    Note: Do not forget to trigger an update on the chart once all your transformations are done

    Related Pages:

      https://www.chartjs.org/docs/latest/developers/updates.html

    Attributes:
    ----------
    :param names: List. Optional. The series names to be removed from the chart. If none all series will be removed
    """
    if names is None:
      return JsObjects.JsVoid('%(varName)s.data.labels = []; %(varName)s.data.datasets = []' % {'varName': self.varName})

    names = JsUtils.jsConvertData(names, None)
    return JsObjects.JsVoid('''
      var indices = [];
      %(varName)s.data.datasets.forEach(function(d, i){ if(%(names)s.includes(d.label)){ indices.push(i)}});
      indices.sort().reverse().forEach(function(i){%(varName)s.data.datasets.splice(i, 1)})
      ''' % {'varName': self.varName, 'names': names})

  @property
  def label(self):
    """
    Description:
    -----------

    """
    return JsObjects.JsString.JsString("%s.data.labels[activePoints[0]['_index']]" % self.varName, isPyData=False)

  @property
  def content(self):
    """
    Description:
    -----------

    """
    return JsObjects.JsObject.JsObject("%s.data.datasets[0].data[activePoints[0]['_index']]" % self.varName, isPyData=False)

  @property
  def value(self):
    """
    Description:
    -----------

    """
    return JsObjects.JsString.JsString("{%(htmlCode)s: {point: %(chart)s.data.datasets[0].data[activePoints[0]['_index']], label: %(chart)s.data.labels[activePoints[0]['_index']]}}" % {'htmlCode': self.src.htmlCode, "chart": self.varName}, isPyData=False)

  def update(self, config=None):
    """
    Description:
    ------------
    Triggers an update of the chart. This can be safely called after updating the data object.

    Related Pages:

			https://www.chartjs.org/docs/latest/developers/api.html

    Attributes:
    ----------
    :param config: A Python dictionary. A config object can be provided with additional configuration for the update proces
    """
    if config is None:
      return JsObjects.JsObject.JsObject("%s.update()" % self.toStr())

    config = JsUtils.jsConvertData(config, None)
    return JsObjects.JsObject.JsObject("%s.update(%s)" % (self.toStr(), config))

  def reset(self):
    """
    Description:
    ------------
    Reset the chart to it's state before the initial animation.
    A new animation can then be triggered using update.

    Usage::

      myLineChart.reset()

    Related Pages:

			https://www.chartjs.org/docs/latest/developers/api.html
    """
    return JsObjects.JsObject.JsObject("%s.reset()" % self.toStr())

  def stop(self):
    """
    Description:
    ------------
    Use this to stop any current animation loop. This will pause the chart during any current animation frame.

    Usage::

      chart.stop()

    Related Pages:

			https://www.chartjs.org/docs/latest/developers/api.html

    :return: 'this' for chainability
    """
    self._js.append("stop()")
    return self

  def resize(self):
    """
    Description:
    ------------
    Use this to manually resize the canvas element.

    This is run each time the canvas container is resized, but you can call this method manually if you change the size of the canvas nodes container element.

    Related Pages:

			https://www.chartjs.org/docs/latest/developers/api.html

    :return: 'this' for chainability
    """
    self._js.append("%s.resize()" % self.toStr())
    return self

  def clear(self):
    """
    Description:
    ------------
    Will clear the chart canvas. Used extensively internally between animation frames, but you might find it useful.

    Usage::

      chart.clear()

    Related Pages:

			https://www.chartjs.org/docs/latest/developers/api.html

    :return: 'this' for chainability
    """
    self._js.append("%s.clear()" % self.toStr())
    return self

  def toBase64Image(self):
    """
    Description:
    ------------
    This returns a base 64 encoded string of the chart in it's current state.

    Related Pages:

			https://www.chartjs.org/docs/latest/developers/api.html

    :return: png data url of the image on the canvas
    """
    return JsObjects.JsObject.JsObject("%s.toBase64Image()" % self.toStr())

  def generateLegend(self):
    """
    Description:
    ------------
    Returns an HTML string of a legend for that chart. The legend is generated from the legendCallback in the options.

    Related Pages:

			https://www.chartjs.org/docs/latest/developers/api.html

    :return: HTML string of a legend for this chart
    """
    return JsObjects.JsObject.JsObject("%s.generateLegend()" % self.toStr())

  def getElementAtEvent(self, jsEvent):
    """
    Description:
    ------------
    Calling getElementAtEvent(event) on your Chart instance passing an argument of an event, or jQuery event, will return the single element at the event position

    Related Pages:

			https://www.chartjs.org/docs/latest/developers/api.html

    Attributes:
    ----------
    :param jsEvent:

    :return: the first element at the event point.
    """
    jsEvent = JsUtils.jsConvertData(jsEvent, None)
    return JsObjects.JsObject.JsObject("%s.getElementAtEvent(%s)" % (self.toStr(), jsEvent))

  def getElementsAtEvent(self, jsEvent):
    """
    Description:
    ------------
    Looks for the element under the event point, then returns all elements at the same data index.

    Related Pages:

			https://www.chartjs.org/docs/latest/developers/api.html

    Attributes:
    ----------
    :param jsEvent:
    """
    jsEvent = JsUtils.jsConvertData(jsEvent, None)
    return JsObjects.JsObject.JsObject("%s.getElementsAtEvent(%s)" % (self.toStr(), jsEvent))

  def getDatasetAtEvent(self, jsEvent):
    """
    Description:
    ------------
    Looks for the element under the event point, then returns all elements from that dataset.

    Related Pages:

			https://www.chartjs.org/docs/latest/developers/api.html

    Attributes:
    ----------
    :param jsEvent:

    :return: an array of elements
    """
    jsEvent = JsUtils.jsConvertData(jsEvent, None)
    return JsObjects.JsArray.JsArray("%s.getDatasetAtEvent(%s)" % (self.toStr(), jsEvent))

  def getDatasetMeta(self, index):
    """
    Description:
    ------------
    Looks for the dataset that matches the current index and returns that metadata.

    Usage::

      chart.getDatasetMeta(0)

    Related Pages:

			https://www.chartjs.org/docs/latest/developers/api.html

    Attributes:
    ----------
    :param index:
    """
    return JsObjects.JsObject.JsObject("%s.getDatasetMeta(%s)" % (self.toStr(), index))

  def render(self, config=None):
    """
    Description:
    ------------
    Triggers a redraw of all chart elements. Note, this does not update elements for new data. Use .update() in that case.

    Usage::

      chart.render({duration: 800, lazy: false, easing: 'easeOutBounce'})

    Related Pages:

			https://www.chartjs.org/docs/latest/developers/api.html

    Attributes:
    ----------
    :param config: A python dictionary as config object
    """
    if config is None:
      return JsObjects.JsObject.JsObject("%s.render()" % self.toStr())

    config = JsUtils.jsConvertData(config, None)
    return JsObjects.JsObject.JsObject("%s.render(%s)" % (self.toStr(), config))

  def destroy(self):
    """
    Description:
    ------------
    Use this to destroy any chart instances that are created.

    Related Pages:

			https://www.chartjs.org/docs/latest/developers/api.html
    """
    return JsObjects.JsObject.JsObject("%s.destroy()" % self.toStr())

  def toStr(self):
    """
    Javascript representation

    :return: Return the Javascript String
    """
    if self._selector is None:
      raise Exception("Selector not defined, use this() or new() first")

    if self.setVar:
      if len(self._js) > 0:
        strData = ".".join(self._js)
        strData = "var %s = %s.%s" % (self.varName, self._selector, strData)
      else:
        strData = "var %s = %s" % (self.varName, self._selector)
      self.setVar = False
    else:
      if len(self._js) > 0:
        strData = ".".join(self._js)
        strData = "%s.%s" % (self.varName, strData)
      else:
        strData = self.varName
    self._js = [] # empty the stack
    return strData


class ChartJsOptTicks(DataAttrs):
  def beginAtZero(self, flag):
    self._attrs["beginAtZero"] = JsUtils.jsConvertData(flag, None)
    return self

  def display(self, flag):
    self._attrs["display"] = JsUtils.jsConvertData(flag, None)
    return self


class ChartJsOptGridLines(DataAttrs):

  def display(self, flag=False):
    """
    If false, do not display grid lines for this axis.

    :param flag:
    :return:
    """
    self._attrs["display"] = JsUtils.jsConvertData(flag, None)
    return self

  def circular(self, flag=False):
    """
    If true, gridlines are circular (on radar chart only).

    :param flag:
    :return:
    """
    self._attrs["circular"] = JsUtils.jsConvertData(flag, None)
    return self

  def color(self, colors=""):
    """
    The color of the grid lines.
    If specified as an array, the first color applies to the first grid line, the second to the second grid line and so on.

    :param colors:
    :return:
    """
    self._attrs["color"] = JsUtils.jsConvertData(colors, None)
    return self

  def borderDash(self, narray):
    """

    :param narray:
    :return:
    """
    self._attrs["borderDash"] = JsUtils.jsConvertData(narray, None)
    return self

  def borderDashOffset(self, n=0.0):
    """
    Offset for line dashes

    :param n:
    :return:
    """
    self._attrs["borderDashOffset"] = JsUtils.jsConvertData(n, None)
    return self

  def lineWidth(self, num):
    """
    Stroke width of grid lines.

    :param num:
    :return:
    """
    self._attrs["lineWidth"] = JsUtils.jsConvertData(num, None)
    return self

  def drawBorder(self, flag=True):
    """
    If true, draw border at the edge between the axis and the chart area.

    :param flag:
    :return:
    """
    self._attrs["drawBorder"] = JsUtils.jsConvertData(flag, None)
    return self

  def drawOnChartArea(self, flag=True):
    """
    If true, draw lines on the chart area inside the axis lines.
    This is useful when there are multiple axes and you need to control which grid lines are drawn.

    :param flag:
    :return:
    """
    self._attrs["drawOnChartArea"] = JsUtils.jsConvertData(flag, None)
    return self

  def drawTicks(self, flag=True):
    """
    If true, draw lines beside the ticks in the axis area beside the chart.

    :param flag:
    :return:
    """
    self._attrs["drawTicks"] = JsUtils.jsConvertData(flag, None)
    return self

  def tickMarkLength(self, n=10):
    """
    Length in pixels that the grid lines will draw into the axis area.

    :param n:
    :return:
    """
    self._attrs["tickMarkLength"] = JsUtils.jsConvertData(n, None)
    return self

  def zeroLineWidth(self, n=1):
    """
    Stroke width of the grid line for the first index (index 0).

    :param n:
    :return:
    """
    self._attrs["zeroLineWidth"] = JsUtils.jsConvertData(n, None)
    return self

  def zeroLineColor(self, color="rgba(0, 0, 0, 0.25)"):
    """
    Stroke color of the grid line for the first index (index 0).

    :param color:
    :return:
    """
    self._attrs["zeroLineColor"] = JsUtils.jsConvertData(color, None)
    return self

  def zeroLineBorderDash(self, narray):
    """
    Length and spacing of dashes of the grid line for the first index (index 0)

    :param narray:
    :return:
    """
    self._attrs["zeroLineBorderDash"] = JsUtils.jsConvertData(narray, None)
    return self

  def zeroLineBorderDashOffset(self, n=0.0):
    """
    Offset for line dashes of the grid line for the first index (index 0).

    :param n:
    :return:
    """
    self._attrs["zeroLineBorderDashOffset"] = JsUtils.jsConvertData(n, None)
    return self

  def offsetGridLines(self, flag=True):
    """
    If true, the bars for a particular data point fall between the grid lines.
    The grid line will move to the left by one half of the tick interval, which is the space between the grid lines.
    If false, the grid line will go right down the middle of the bars.
    This is set to true for a category scale in a bar chart while false for other scales or chart types by default.

    :param flag:
    :return:
    """
    self._attrs["offsetGridLines"] = JsUtils.jsConvertData(flag, None)
    return self


class ChartJsOptScale(DataAttrs):

  def ticks(self):
    if not "ticks" in self._attrs:
      self._attrs["ticks"] = ChartJsOptTicks(self._report)
    return self._attrs["ticks"]

  @property
  def gridLines(self):
    """

    :rtype: ChartJsOptGridLines
    """
    if not "gridLines" in self._attrs:
      self._attrs["gridLines"] = ChartJsOptGridLines(self._report)
    return self._attrs["gridLines"]

  def stacked(self, flag=False):
    pass


class ChartJsOptScaleBar(ChartJsOptScale):
  def barPercentage(self, n=0.9):
    """
    Percent (0-1) of the available width each bar should be within the category width.
    1.0 will take the whole category width and put the bars right next to each other

    :param n:

    :return:
    """
    if n > 1:
      raise Exception("n cannot exceed 1")

    self._attrs["barPercentage"] = JsUtils.jsConvertData(n, None)
    return self

  def categoryPercentage(self, n=0.8):
    """
    Percent (0-1) of the available width each category should be within the sample width.

    :param n:
    :return:
    """
    self._attrs["categoryPercentage"] = JsUtils.jsConvertData(n, None)
    return self

  def barThickness(self, width="flex"):
    """
    Manually set width of each bar in pixels.

    If set to 'flex', it computes "optimal" sample widths that globally arrange bars side by side.
    If not set (default), bars are equally sized based on the smallest interval.

    :return:
    """
    self._attrs["barThickness"] = JsUtils.jsConvertData(width, None)
    return self

  def maxBarThickness(self, width):
    """
    Set this to ensure that bars are not sized thicker than this.

    :param width: A float
    :return:
    """
    self._attrs["barThickness"] = JsUtils.jsConvertData(width, None)
    return self

  def minBarLength(self, width):
    """

    :param width:
    :return:
    """
    self._attrs["minBarLength"] = JsUtils.jsConvertData(width, None)
    return self


class ChartJsOptTooltip(DataAttrs):
  pass


class ChartJsOptLabels(DataAttrs):
  @property
  def fontColor(self):
    """
    https://www.chartjs.org/docs/latest/general/fonts.html

    """
    return self._attrs["fontColor"]

  @fontColor.setter
  def fontColor(self, val):
    self._attrs["fontColor"] = val


class ChartJsOptPadding(DataAttrs):
  @property
  def left(self):
    """
    https://www.chartjs.org/docs/latest/general/fonts.html

    """
    return self._attrs["left"]

  @left.setter
  def left(self, val):
    self._attrs["left"] = val

  @property
  def right(self):
    """
    https://www.chartjs.org/docs/latest/configuration/layout.html

    """
    return self._attrs["right"]

  @right.setter
  def right(self, val):
    self._attrs["right"] = val

  @property
  def top(self):
    """
    https://www.chartjs.org/docs/latest/configuration/layout.html

    """
    return self._attrs["top"]

  @top.setter
  def top(self, val):
    self._attrs["top"] = val

  @property
  def bottom(self):
    """
    https://www.chartjs.org/docs/latest/configuration/layout.html

    """
    return self._attrs["bottom"]

  @bottom.setter
  def bottom(self, val):
    self._attrs["bottom"] = val


class OptionsLegend(DataAttrs):
  def display(self, flag=True):
    """
    Is the legend shown?

    Related Pages:

			https://www.chartjs.org/docs/latest/configuration/legend.html

    :param flag:
    :return:
    """
    self._attrs["display"] = JsUtils.jsConvertData(flag, None)
    return self

  def position(self, location="top"):
    """
    Position of the legend

    :param location:
    :return:
    """
    self._attrs["position"] = JsUtils.jsConvertData(location, None)
    return self

  def fullWidth(self, flag=True):
    """

    :param flag:
    :return:
    """

  def onClick(self, callback):
    """
    A callback that is called when a click event is registered on a label item.

    :param callback:
    :return:
    """
    self._attrs["onClick"] = JsUtils.jsConvertFncs(callback)
    return self

  def onHover(self, callback):
    """
    A callback that is called when a 'mousemove' event is registered on top of a label item.

    :param callback:
    :return:
    """
    self._attrs["onHover"] = JsUtils.jsConvertFncs(callback)
    return self

  def onLeave(self, callback):
    """

    :param callback:
    :return:
    """
    self._attrs["onLeave"] = JsUtils.jsConvertFncs(callback)
    return self

  def reverse(self, flag=False):
    """
    Legend will show datasets in reverse order.

    :param flag:
    :return:
    """
    self._attrs["reverse"] = JsUtils.jsConvertFncs(flag)
    return self

  @property
  def labels(self):
    """

    :rtype: ChartJsOptLabels
    :return:
    """
    if not 'labels' in self._attrs:
      self._attrs["labels"] = ChartJsOptLabels(self._report)
    return self._attrs["labels"]


class OptionsTitle(DataAttrs):
  @property
  def display(self):
    """
    https://www.chartjs.org/docs/latest/configuration/title.html

    """
    return self._attrs["display"]

  @display.setter
  def display(self, val):
    self._attrs["display"] = val

  @property
  def text(self):
    return self._attrs["text"]

  @text.setter
  def text(self, val):
    self._attrs["text"] = val

  @property
  def position(self):
    return self._attrs["position"]

  @position.setter
  def position(self, val):
    self._attrs["position"] = val

  @property
  def fontSize(self):
    return self._attrs["fontSize"]

  @fontSize.setter
  def fontSize(self, val):
    self._attrs["fontSize"] = val

  @property
  def fontFamily(self):
    return self._attrs["fontFamily"]

  @fontFamily.setter
  def fontFamily(self, val):
    self._attrs["fontFamily"] = val

  @property
  def fontColor(self):
    return self._attrs["fontColor"]

  @fontColor.setter
  def fontColor(self, val):
    self._attrs["fontColor"] = val

  @property
  def fontStyle(self):
    return self._attrs["fontStyle"]

  @fontStyle.setter
  def fontStyle(self, val):
    self._attrs["fontStyle"] = val

  @property
  def padding(self):
    return self._attrs["padding"]

  @padding.setter
  def padding(self, val):
    self._attrs["padding"] = val

  @property
  def lineHeight(self):
    return self._attrs["lineHeight"]

  @lineHeight.setter
  def lineHeight(self, val):
    self._attrs["lineHeight"] = val


class Options(DataAttrs):

  def __init__(self, report, attrs=None, oprions=None):
    super(Options, self).__init__(report, attrs, oprions)

  def title(self):
    if self._attrs.get("title") is None:
      self._attrs['title'] = OptionsTitle(self._report)
    return self._attrs['title']


class DataSetPie(DataAttrs):

  @property
  def backgroundColor(self):
    """
    Arc background color.

    https://www.chartjs.org/docs/latest/charts/doughnut.html
    """
    return self._attrs["backgroundColor"]

  @backgroundColor.setter
  def backgroundColor(self, val):
    self._attrs["backgroundColor"] = val

  @property
  def borderAlign(self):
    """
    https://www.chartjs.org/docs/latest/charts/doughnut.html
    """
    return self._attrs["borderAlign"]

  @borderAlign.setter
  def borderAlign(self, val):
    self._attrs["borderAlign"] = val

  @property
  def borderColor(self):
    """
    Arc border color.

    https://www.chartjs.org/docs/latest/charts/doughnut.html
    """
    return self._attrs["borderColor"]

  @borderColor.setter
  def borderColor(self, val):
    self._attrs["borderColor"] = val

  @property
  def borderWidth(self):
    """
    Arc border width (in pixels).

    https://www.chartjs.org/docs/latest/charts/doughnut.html
    """
    return self._attrs["borderWidth"]

  @borderWidth.setter
  def borderWidth(self, val):
    self._attrs["borderWidth"] = val

  @property
  def data(self):
    """
    https://www.chartjs.org/docs/latest/charts/doughnut.html
    """
    return self._attrs["data"]

  @data.setter
  def data(self, val):
    self._attrs["data"] = val

  @property
  def fillOpacity(self):
    """
    Convert the hexadecimal color to the corresponding RGB one with the opacity

    https://www.chartjs.org/docs/latest/charts/doughnut.html
    """
    return self._attrs["backgroundColor"]

  @fillOpacity.setter
  def fillOpacity(self, val):
    bgColors = self._attrs["backgroundColor"]
    if isinstance(bgColors, list):
      opColors = []
      for c in bgColors:
        color = Colors.getHexToRgb(c)
        opColors.append("rgba(%s, %s, %s, %s)" % (color[0], color[1], color[2], val))
      self._attrs["backgroundColor"] = opColors
    else:
      color = Colors.getHexToRgb(self._attrs["backgroundColor"])
      self._attrs["backgroundColor"] = "rgba(%s, %s, %s, %s)" % (color[0], color[1], color[2], val)

  @property
  def hoverBackgroundColor(self):
    """
    Arc background color when hovered.

    https://www.chartjs.org/docs/latest/charts/doughnut.html
    """
    return self._attrs["hoverBackgroundColor"]

  @hoverBackgroundColor.setter
  def hoverBackgroundColor(self, val):
    self._attrs["hoverBackgroundColor"] = val

  @property
  def hoverBorderColor(self):
    """
    Arc border color when hovered.

    https://www.chartjs.org/docs/latest/charts/doughnut.html
    """
    return self._attrs["hoverBorderColor"]

  @hoverBorderColor.setter
  def hoverBorderColor(self, val):
    self._attrs["hoverBorderColor"] = val

  @property
  def hoverBorderWidth(self):
    """
    Arc border width when hovered (in pixels).

    https://www.chartjs.org/docs/latest/charts/doughnut.html
    """
    return self._attrs["hoverBorderWidth"]

  @hoverBorderWidth.setter
  def hoverBorderWidth(self, val):
    self._attrs["hoverBorderWidth"] = val

  @property
  def weight(self):
    """
    The relative thickness of the dataset. Providing a value for weight will cause the pie or doughnut dataset to be drawn with a thickness relative to the sum of all the dataset weight values.

    https://www.chartjs.org/docs/latest/charts/doughnut.html
    """
    return self._attrs["weight"]

  @weight.setter
  def weight(self, val):
    self._attrs["weight"] = val


class DataSetScatterLine(DataAttrs):
  @property
  def backgroundColor(self):
    """
    https://www.chartjs.org/docs/latest/charts/line.html
    """
    return self._attrs["backgroundColor"]

  @backgroundColor.setter
  def backgroundColor(self, val):
    self._attrs["backgroundColor"] = val

  @property
  def borderCapStyle(self):
    """
    https://www.chartjs.org/docs/latest/charts/line.html
    """
    return self._attrs["borderCapStyle"]

  @borderCapStyle.setter
  def borderCapStyle(self, val):
    self._attrs["borderCapStyle"] = val

  @property
  def borderColor(self):
    """
    https://www.chartjs.org/docs/latest/charts/line.html
    """
    return self._attrs["borderColor"]

  @borderColor.setter
  def borderColor(self, val):
    self._attrs["borderColor"] = val

  @property
  def borderDash(self):
    """
    https://www.chartjs.org/docs/latest/charts/line.html
    """
    return self._attrs["borderDash"]

  @borderDash.setter
  def borderDash(self, val):
    self._attrs["borderDash"] = val

  @property
  def borderDashOffset(self):
    """
    https://www.chartjs.org/docs/latest/charts/line.html
    """
    return self._attrs["borderDashOffset"]

  @borderDashOffset.setter
  def borderDashOffset(self, val):
    self._attrs["borderDashOffset"] = val

  @property
  def borderJoinStyle(self):
    """
    https://www.chartjs.org/docs/latest/charts/line.html
    """
    return self._attrs["borderJoinStyle"]

  @borderJoinStyle.setter
  def borderJoinStyle(self, val):
    self._attrs["borderJoinStyle"] = val

  @property
  def borderWidth(self):
    """
    https://www.chartjs.org/docs/latest/charts/line.html
    """
    return self._attrs["borderWidth"]

  @borderWidth.setter
  def borderWidth(self, val):
    self._attrs["borderWidth"] = val

  @property
  def cubicInterpolationMode(self):
    """
    https://www.chartjs.org/docs/latest/charts/line.html
    """
    return self._attrs["cubicInterpolationMode"]

  @cubicInterpolationMode.setter
  def cubicInterpolationMode(self, val):
    self._attrs["cubicInterpolationMode"] = val

  @property
  def clip(self):
    """
    https://www.chartjs.org/docs/latest/charts/line.html
    """
    return self._attrs["clip"]

  @clip.setter
  def clip(self, val):
    self._attrs["clip"] = val

  @property
  def fill(self):
    """
    https://www.chartjs.org/docs/latest/charts/line.html
    """
    return self._attrs["fill"]

  @fill.setter
  def fill(self, val):
    self._attrs["fill"] = val

  @property
  def fillOpacity(self):
    """
    Convert the hexadecimal color to the corresponding RGB one with the opacity

    https://www.chartjs.org/docs/latest/charts/doughnut.html
    """
    return self._attrs["backgroundColor"]

  @fillOpacity.setter
  def fillOpacity(self, val):
    bgColors = self._attrs["backgroundColor"]
    if isinstance(bgColors, list):
      opColors = []
      for c in bgColors:
        color = Colors.getHexToRgb(c)
        opColors.append("rgba(%s, %s, %s, %s)" % (color[0], color[1], color[2], val))
      self._attrs["backgroundColor"] = opColors
    else:
      color = Colors.getHexToRgb(self._attrs["backgroundColor"])
      self._attrs["backgroundColor"] = "rgba(%s, %s, %s, %s)" % (color[0], color[1], color[2], val)

  @property
  def hoverBackgroundColor(self):
    """
    https://www.chartjs.org/docs/latest/charts/line.html
    """
    return self._attrs["hoverBackgroundColor"]

  @hoverBackgroundColor.setter
  def hoverBackgroundColor(self, val):
    self._attrs["hoverBackgroundColor"] = val

  @property
  def hoverBorderCapStyle(self):
    """
    https://www.chartjs.org/docs/latest/charts/line.html
    """
    return self._attrs["hoverBorderCapStyle"]

  @hoverBorderCapStyle.setter
  def hoverBorderCapStyle(self, val):
    self._attrs["hoverBorderCapStyle"] = val

  @property
  def hoverBorderColor(self):
    """
    https://www.chartjs.org/docs/latest/charts/line.html
    """
    return self._attrs["hoverBorderColor"]

  @hoverBorderColor.setter
  def hoverBorderColor(self, val):
    self._attrs["hoverBorderColor"] = val

  @property
  def hoverBorderDash(self):
    """
    https://www.chartjs.org/docs/latest/charts/line.html
    """
    return self._attrs["hoverBorderDash"]

  @hoverBorderDash.setter
  def hoverBorderDash(self, val):
    self._attrs["hoverBorderDash"] = val

  @property
  def hoverBorderDashOffset(self):
    """
    https://www.chartjs.org/docs/latest/charts/line.html
    """
    return self._attrs["hoverBorderDashOffset"]

  @hoverBorderDashOffset.setter
  def hoverBorderDashOffset(self, val):
    self._attrs["hoverBorderDashOffset"] = val

  @property
  def hoverBorderJoinStyle(self):
    """
    https://www.chartjs.org/docs/latest/charts/line.html
    """
    return self._attrs["hoverBorderJoinStyle"]

  @hoverBorderJoinStyle.setter
  def hoverBorderJoinStyle(self, val):
    self._attrs["hoverBorderJoinStyle"] = val

  @property
  def hoverBorderWidth(self):
    """
    https://www.chartjs.org/docs/latest/charts/line.html
    """
    return self._attrs["hoverBorderWidth"]

  @hoverBorderWidth.setter
  def hoverBorderWidth(self, val):
    self._attrs["hoverBorderWidth"] = val

  @property
  def label(self):
    """
    https://www.chartjs.org/docs/latest/charts/line.html
    """
    return self._attrs["label"]

  @label.setter
  def label(self, val):
    self._attrs["label"] = val

  @property
  def lineTension(self):
    """
    https://www.chartjs.org/docs/latest/charts/line.html
    """
    return self._attrs["lineTension"]

  @lineTension.setter
  def lineTension(self, val):
    self._attrs["lineTension"] = val

  @property
  def order(self):
    """
    https://www.chartjs.org/docs/latest/charts/line.html
    """
    return self._attrs["order"]

  @order.setter
  def order(self, val):
    self._attrs["order"] = val

  @property
  def pointBackgroundColor(self):
    """
    https://www.chartjs.org/docs/latest/charts/line.html
    """
    return self._attrs["pointBackgroundColor"]

  @pointBackgroundColor.setter
  def pointBackgroundColor(self, val):
    self._attrs["pointBackgroundColor"] = val

  @property
  def pointBorderColor(self):
    """
    https://www.chartjs.org/docs/latest/charts/line.html
    """
    return self._attrs["pointBorderColor"]

  @pointBorderColor.setter
  def pointBorderColor(self, val):
    self._attrs["pointBorderColor"] = val

  @property
  def pointBorderWidth(self):
    """
    https://www.chartjs.org/docs/latest/charts/line.html
    """
    return self._attrs["pointBorderWidth"]

  @pointBorderWidth.setter
  def pointBorderWidth(self, val):
    self._attrs["pointBorderWidth"] = val

  @property
  def pointHitRadius(self):
    """
    https://www.chartjs.org/docs/latest/charts/line.html
    """
    return self._attrs["pointHitRadius"]

  @pointHitRadius.setter
  def pointHitRadius(self, val):
    self._attrs["pointHitRadius"] = val

  @property
  def pointHoverBackgroundColor(self):
    """
    https://www.chartjs.org/docs/latest/charts/line.html
    """
    return self._attrs["pointHoverBackgroundColor"]

  @pointHoverBackgroundColor.setter
  def pointHoverBackgroundColor(self, val):
    self._attrs["pointHoverBackgroundColor"] = val

  @property
  def pointHoverBorderColor(self):
    """
    https://www.chartjs.org/docs/latest/charts/line.html
    """
    return self._attrs["pointHoverBorderColor"]

  @pointHoverBorderColor.setter
  def pointHoverBorderColor(self, val):
    self._attrs["pointHoverBorderColor"] = val

  @property
  def pointHoverBorderWidth(self):
    """
    https://www.chartjs.org/docs/latest/charts/line.html
    """
    return self._attrs["pointHoverBorderWidth"]

  @pointHoverBorderWidth.setter
  def pointHoverBorderWidth(self, val):
    self._attrs["pointHoverBorderWidth"] = val

  @property
  def pointHoverRadius(self):
    """
    https://www.chartjs.org/docs/latest/charts/line.html
    """
    return self._attrs["pointHoverRadius"]

  @pointHoverRadius.setter
  def pointHoverRadius(self, val):
    self._attrs["pointHoverRadius"] = val

  @property
  def pointRadius(self):
    """
    https://www.chartjs.org/docs/latest/charts/line.html
    """
    return self._attrs["pointRadius"]

  @pointRadius.setter
  def pointRadius(self, val):
    self._attrs["pointRadius"] = val

  @property
  def pointRotation(self):
    """
    https://www.chartjs.org/docs/latest/charts/line.html
    """
    return self._attrs["pointRotation"]

  @pointRotation.setter
  def pointRotation(self, val):
    self._attrs["pointRotation"] = val

  @property
  def pointStyle(self):
    """
    https://www.chartjs.org/docs/latest/charts/line.html
    """
    return self._attrs["pointStyle"]

  @pointStyle.setter
  def pointStyle(self, val):
    self._attrs["pointStyle"] = val

  @property
  def showLine(self):
    """
    https://www.chartjs.org/docs/latest/charts/line.html
    """
    return self._attrs["showLine"]

  @showLine.setter
  def showLine(self, val):
    self._attrs["showLine"] = val

  @property
  def spanGaps(self):
    """
    https://www.chartjs.org/docs/latest/charts/line.html
    """
    return self._attrs["spanGaps"]

  @spanGaps.setter
  def spanGaps(self, val):
    self._attrs["spanGaps"] = val

  @property
  def steppedLine(self):
    """
    https://www.chartjs.org/docs/latest/charts/line.html
    """
    return self._attrs["steppedLine"]

  @steppedLine.setter
  def steppedLine(self, val):
    self._attrs["steppedLine"] = val

  @property
  def xAxisID(self):
    """
    https://www.chartjs.org/docs/latest/charts/line.html
    """
    return self._attrs["xAxisID"]

  @xAxisID.setter
  def xAxisID(self, val):
    self._attrs["xAxisID"] = val

  @property
  def yAxisID(self):
    """
    https://www.chartjs.org/docs/latest/charts/line.html
    """
    return self._attrs["yAxisID"]

  @yAxisID.setter
  def yAxisID(self, val):
    self._attrs["yAxisID"] = val


class DataSetBar(DataAttrs):
  @property
  def backgroundColor(self):
    """
    https://www.chartjs.org/docs/latest/charts/bar.html
    """
    return self._attrs["backgroundColor"]

  @backgroundColor.setter
  def backgroundColor(self, val):
    self._attrs["backgroundColor"] = val

  @property
  def barPercentage(self):
    """
    https://www.chartjs.org/docs/latest/charts/bar.html
    """
    return self._attrs["barPercentage"]

  @barPercentage.setter
  def barPercentage(self, val):
    self._attrs["barPercentage"] = val

  @property
  def barThickness(self):
    """
    https://www.chartjs.org/docs/latest/charts/bar.html
    """
    return self._attrs["barThickness"]

  @barThickness.setter
  def barThickness(self, val):
    self._attrs["barThickness"] = val

  @property
  def fillOpacity(self):
    """
    Convert the hexadecimal color to the corresponding RGB one with the opacity

    https://www.chartjs.org/docs/latest/charts/doughnut.html
    """
    return self._attrs["backgroundColor"]

  @fillOpacity.setter
  def fillOpacity(self, val):
    bgColors = self._attrs["backgroundColor"]
    if isinstance(bgColors, list):
      opColors = []
      for c in bgColors:
        color = Colors.getHexToRgb(c)
        opColors.append("rgba(%s, %s, %s, %s)" % (color[0], color[1], color[2], val))
      self._attrs["backgroundColor"] = opColors
    else:
      color = Colors.getHexToRgb(self._attrs["backgroundColor"])
      self._attrs["backgroundColor"] = "rgba(%s, %s, %s, %s)" % (color[0], color[1], color[2], val)

  @property
  def maxBarThickness(self):
    """
    https://www.chartjs.org/docs/latest/charts/bar.html
    """
    return self._attrs["maxBarThickness"]

  @maxBarThickness.setter
  def maxBarThickness(self, val):
    self._attrs["maxBarThickness"] = val

  @property
  def minBarLength(self):
    """
    https://www.chartjs.org/docs/latest/charts/bar.html
    """
    return self._attrs["minBarLength"]

  @minBarLength.setter
  def minBarLength(self, val):
    self._attrs["minBarLength"] = val

  @property
  def borderColor(self):
    """
    https://www.chartjs.org/docs/latest/charts/bar.html
    """
    return self._attrs["borderColor"]

  @borderColor.setter
  def borderColor(self, val):
    self._attrs["borderColor"] = val

  @property
  def borderSkipped(self):
    """
    https://www.chartjs.org/docs/latest/charts/bar.html
    """
    return self._attrs["borderSkipped"]

  @borderSkipped.setter
  def borderSkipped(self, val):
    self._attrs["borderSkipped"] = val

  @property
  def borderWidth(self):
    """
    https://www.chartjs.org/docs/latest/charts/bar.html
    """
    return self._attrs["borderWidth"]

  @borderWidth.setter
  def borderWidth(self, val):
    self._attrs["borderWidth"] = val

  @property
  def data(self):
    """
    https://www.chartjs.org/docs/latest/charts/bar.html
    """
    return self._attrs["data"]

  @data.setter
  def data(self, val):
    self._attrs["data"] = val

  @property
  def hoverBackgroundColor(self):
    """
    https://www.chartjs.org/docs/latest/charts/bar.html
    """
    return self._attrs["hoverBackgroundColor"]

  @hoverBackgroundColor.setter
  def hoverBackgroundColor(self, val):
    self._attrs["hoverBackgroundColor"] = val

  @property
  def hoverBorderColor(self):
    """
    https://www.chartjs.org/docs/latest/charts/bar.html
    """
    return self._attrs["hoverBorderColor"]

  @hoverBorderColor.setter
  def hoverBorderColor(self, val):
    self._attrs["hoverBorderColor"] = val

  @property
  def hoverBorderWidth(self):
    """
    https://www.chartjs.org/docs/latest/charts/bar.html
    """
    return self._attrs["hoverBorderWidth"]

  @hoverBorderWidth.setter
  def hoverBorderWidth(self, val):
    self._attrs["hoverBorderWidth"] = val

  @property
  def label(self):
    """
    https://www.chartjs.org/docs/latest/charts/bar.html
    """
    return self._attrs["label"]

  @label.setter
  def label(self, val):
    self._attrs["label"] = val

  @property
  def order(self):
    """
    https://www.chartjs.org/docs/latest/charts/bar.html
    """
    return self._attrs["order"]

  @order.setter
  def order(self, val):
    self._attrs["order"] = val

  @property
  def xAxisID(self):
    """
    https://www.chartjs.org/docs/latest/charts/bar.html
    """
    return self._attrs["xAxisID"]

  @xAxisID.setter
  def xAxisID(self, val):
    self._attrs["xAxisID"] = val

  @property
  def yAxisID(self):
    """
    https://www.chartjs.org/docs/latest/charts/bar.html
    """
    return self._attrs["yAxisID"]

  @yAxisID.setter
  def yAxisID(self, val):
    self._attrs["yAxisID"] = val


class DataSetPolar(DataAttrs):
  @property
  def backgroundColor(self):
    """
    https://www.chartjs.org/docs/latest/charts/polar.html
    """
    return self._attrs["backgroundColor"]

  @backgroundColor.setter
  def backgroundColor(self, val):
    self._attrs["backgroundColor"] = val

  @property
  def borderAlign(self):
    """
    https://www.chartjs.org/docs/latest/charts/polar.html
    """
    return self._attrs["borderAlign"]

  @borderAlign.setter
  def borderAlign(self, val):
    self._attrs["borderAlign"] = val

  @property
  def borderColor(self):
    """
    https://www.chartjs.org/docs/latest/charts/polar.html
    """
    return self._attrs["borderColor"]

  @borderColor.setter
  def borderColor(self, val):
    self._attrs["borderColor"] = val

  @property
  def borderWidth(self):
    """
    https://www.chartjs.org/docs/latest/charts/polar.html
    """
    return self._attrs["borderWidth"]

  @borderWidth.setter
  def borderWidth(self, val):
    self._attrs["borderWidth"] = val

  @property
  def data(self):
    """
    https://www.chartjs.org/docs/latest/charts/polar.html
    """
    return self._attrs["data"]

  @data.setter
  def data(self, val):
    self._attrs["data"] = val

  @property
  def fillOpacity(self):
    """
    Convert the hexadecimal color to the corresponding RGB one with the opacity

    https://www.chartjs.org/docs/latest/charts/doughnut.html
    """
    return self._attrs["backgroundColor"]

  @fillOpacity.setter
  def fillOpacity(self, val):
    bgColors = self._attrs["backgroundColor"]
    if isinstance(bgColors, list):
      opColors = []
      for c in bgColors:
        color = Colors.getHexToRgb(c)
        opColors.append("rgba(%s, %s, %s, %s)" % (color[0], color[1], color[2], val))
      self._attrs["backgroundColor"] = opColors
    else:
      color = Colors.getHexToRgb(self._attrs["backgroundColor"])
      self._attrs["backgroundColor"] = "rgba(%s, %s, %s, %s)" % (color[0], color[1], color[2], val)

  @property
  def hoverBackgroundColor(self):
    """
    https://www.chartjs.org/docs/latest/charts/polar.html
    """
    return self._attrs["hoverBackgroundColor"]

  @hoverBackgroundColor.setter
  def hoverBackgroundColor(self, val):
    self._attrs["hoverBackgroundColor"] = val

  @property
  def hoverBorderColor(self):
    """
    https://www.chartjs.org/docs/latest/charts/polar.html
    """
    return self._attrs["hoverBorderColor"]

  @hoverBorderColor.setter
  def hoverBorderColor(self, val):
    self._attrs["hoverBorderColor"] = val

  @property
  def hoverBorderWidth(self):
    """
    https://www.chartjs.org/docs/latest/charts/polar.html
    """
    return self._attrs["hoverBorderWidth"]

  @hoverBorderWidth.setter
  def hoverBorderWidth(self, val):
    self._attrs["hoverBorderWidth"] = val


class DataSetRadar(DataAttrs):
  @property
  def backgroundColor(self):
    """
    https://www.chartjs.org/docs/latest/charts/radar.html
    """
    return self._attrs["backgroundColor"]

  @backgroundColor.setter
  def backgroundColor(self, val):
    self._attrs["backgroundColor"] = val

  @property
  def borderCapStyle(self):
    """
    https://www.chartjs.org/docs/latest/charts/radar.html
    """
    return self._attrs["borderCapStyle"]

  @borderCapStyle.setter
  def borderCapStyle(self, val):
    self._attrs["borderCapStyle"] = val

  @property
  def borderColor(self):
    """
    https://www.chartjs.org/docs/latest/charts/radar.html
    """
    return self._attrs["borderColor"]

  @borderColor.setter
  def borderColor(self, val):
    self._attrs["borderColor"] = val

  @property
  def borderDash(self):
    """
    https://www.chartjs.org/docs/latest/charts/radar.html
    """
    return self._attrs["borderDash"]

  @borderDash.setter
  def borderDash(self, val):
    self._attrs["borderDash"] = val

  @property
  def borderDashOffset(self):
    """
    https://www.chartjs.org/docs/latest/charts/radar.html
    """
    return self._attrs["borderDashOffset"]

  @borderDashOffset.setter
  def borderDashOffset(self, val):
    self._attrs["borderDashOffset"] = val

  @property
  def borderJoinStyle(self):
    """
    https://www.chartjs.org/docs/latest/charts/radar.html
    """
    return self._attrs["borderJoinStyle"]

  @borderJoinStyle.setter
  def borderJoinStyle(self, val):
    self._attrs["borderJoinStyle"] = val

  @property
  def borderWidth(self):
    """
    https://www.chartjs.org/docs/latest/charts/radar.html
    """
    return self._attrs["borderWidth"]

  @borderWidth.setter
  def borderWidth(self, val):
    self._attrs["borderWidth"] = val

  @property
  def hoverBackgroundColor(self):
    """
    https://www.chartjs.org/docs/latest/charts/radar.html
    """
    return self._attrs["hoverBackgroundColor"]

  @hoverBackgroundColor.setter
  def hoverBackgroundColor(self, val):
    self._attrs["hoverBackgroundColor"] = val

  @property
  def hoverBorderCapStyle(self):
    """
    https://www.chartjs.org/docs/latest/charts/radar.html
    """
    return self._attrs["hoverBorderCapStyle"]

  @hoverBorderCapStyle.setter
  def hoverBorderCapStyle(self, val):
    self._attrs["hoverBorderCapStyle"] = val

  @property
  def hoverBorderColor(self):
    """
    https://www.chartjs.org/docs/latest/charts/radar.html
    """
    return self._attrs["hoverBorderColor"]

  @hoverBorderColor.setter
  def hoverBorderColor(self, val):
    self._attrs["hoverBorderColor"] = val

  @property
  def hoverBorderDash(self):
    """
    https://www.chartjs.org/docs/latest/charts/radar.html
    """
    return self._attrs["hoverBorderDash"]

  @hoverBorderDash.setter
  def hoverBorderDash(self, val):
    self._attrs["hoverBorderDash"] = val

  @property
  def hoverBorderDashOffset(self):
    """
    https://www.chartjs.org/docs/latest/charts/radar.html
    """
    return self._attrs["hoverBorderDashOffset"]

  @hoverBorderDashOffset.setter
  def hoverBorderDashOffset(self, val):
    self._attrs["hoverBorderDashOffset"] = val

  @property
  def hoverBorderJoinStyle(self):
    """
    https://www.chartjs.org/docs/latest/charts/radar.html
    """
    return self._attrs["hoverBorderJoinStyle"]

  @hoverBorderJoinStyle.setter
  def hoverBorderJoinStyle(self, val):
    self._attrs["hoverBorderJoinStyle"] = val

  @property
  def hoverBorderWidth(self):
    """
    https://www.chartjs.org/docs/latest/charts/radar.html
    """
    return self._attrs["hoverBorderWidth"]

  @hoverBorderWidth.setter
  def hoverBorderWidth(self, val):
    self._attrs["hoverBorderWidth"] = val

  @property
  def fill(self):
    """
    https://www.chartjs.org/docs/latest/charts/radar.html
    """
    return self._attrs["fill"]

  @fill.setter
  def fill(self, val):
    self._attrs["fill"] = val

  @property
  def fillOpacity(self):
    """
    Convert the hexadecimal color to the corresponding RGB one with the opacity

    https://www.chartjs.org/docs/latest/charts/doughnut.html
    """
    return self._attrs["backgroundColor"]

  @fillOpacity.setter
  def fillOpacity(self, val):
    bgColors = self._attrs["backgroundColor"]
    if isinstance(bgColors, list):
      opColors = []
      for c in bgColors:
        color = Colors.getHexToRgb(c)
        opColors.append("rgba(%s, %s, %s, %s)" % (color[0], color[1], color[2], val))
      self._attrs["backgroundColor"] = opColors
    else:
      color = Colors.getHexToRgb(self._attrs["backgroundColor"])
      self._attrs["backgroundColor"] = "rgba(%s, %s, %s, %s)" % (color[0], color[1], color[2], val)

  @property
  def label(self):
    """
    https://www.chartjs.org/docs/latest/charts/radar.html
    """
    return self._attrs["label"]

  @label.setter
  def label(self, val):
    self._attrs["label"] = val

  @property
  def order(self):
    """
    https://www.chartjs.org/docs/latest/charts/radar.html
    """
    return self._attrs["order"]

  @order.setter
  def order(self, val):
    self._attrs["order"] = val

  @property
  def lineTension(self):
    """
    https://www.chartjs.org/docs/latest/charts/radar.html
    """
    return self._attrs["lineTension"]

  @lineTension.setter
  def lineTension(self, val):
    self._attrs["lineTension"] = val

  @property
  def pointBackgroundColor(self):
    """
    https://www.chartjs.org/docs/latest/charts/radar.html
    """
    return self._attrs["pointBackgroundColor"]

  @pointBackgroundColor.setter
  def pointBackgroundColor(self, val):
    self._attrs["pointBackgroundColor"] = val

  @property
  def pointBorderColor(self):
    """
    https://www.chartjs.org/docs/latest/charts/radar.html
    """
    return self._attrs["pointBorderColor"]

  @pointBorderColor.setter
  def pointBorderColor(self, val):
    self._attrs["pointBorderColor"] = val

  @property
  def pointBorderWidth(self):
    """
    https://www.chartjs.org/docs/latest/charts/radar.html
    """
    return self._attrs["pointBorderWidth"]

  @pointBorderWidth.setter
  def pointBorderWidth(self, val):
    self._attrs["pointBorderWidth"] = val

  @property
  def pointHitRadius(self):
    """
    https://www.chartjs.org/docs/latest/charts/radar.html
    """
    return self._attrs["pointHitRadius"]

  @pointHitRadius.setter
  def pointHitRadius(self, val):
    self._attrs["pointHitRadius"] = val

  @property
  def pointHoverBackgroundColor(self):
    """
    https://www.chartjs.org/docs/latest/charts/radar.html
    """
    return self._attrs["pointHoverBackgroundColor"]

  @pointHoverBackgroundColor.setter
  def pointHoverBackgroundColor(self, val):
    self._attrs["pointHoverBackgroundColor"] = val

  @property
  def pointHoverBorderColor(self):
    """
    https://www.chartjs.org/docs/latest/charts/radar.html
    """
    return self._attrs["pointHoverBorderColor"]

  @pointHoverBorderColor.setter
  def pointHoverBorderColor(self, val):
    self._attrs["pointHoverBorderColor"] = val

  @property
  def pointHoverBorderWidth(self):
    """
    https://www.chartjs.org/docs/latest/charts/radar.html
    """
    return self._attrs["pointHoverBorderWidth"]

  @pointHoverBorderWidth.setter
  def pointHoverBorderWidth(self, val):
    self._attrs["pointHoverBorderWidth"] = val

  @property
  def pointHoverRadius(self):
    """
    https://www.chartjs.org/docs/latest/charts/radar.html
    """
    return self._attrs["pointHoverRadius"]

  @pointHoverRadius.setter
  def pointHoverRadius(self, val):
    self._attrs["pointHoverRadius"] = val

  @property
  def pointRadius(self):
    """
    https://www.chartjs.org/docs/latest/charts/radar.html
    """
    return self._attrs["pointRadius"]

  @pointRadius.setter
  def pointRadius(self, val):
    self._attrs["pointRadius"] = val

  @property
  def pointRotation(self):
    """
    https://www.chartjs.org/docs/latest/charts/radar.html
    """
    return self._attrs["pointRotation"]

  @pointRotation.setter
  def pointRotation(self, val):
    self._attrs["pointRotation"] = val

  @property
  def pointStyle(self):
    """
    https://www.chartjs.org/docs/latest/charts/radar.html
    """
    return self._attrs["pointStyle"]

  @pointStyle.setter
  def pointStyle(self, val):
    self._attrs["pointStyle"] = val

  @property
  def spanGaps(self):
    """
    https://www.chartjs.org/docs/latest/charts/radar.html
    """
    return self._attrs["spanGaps"]

  @spanGaps.setter
  def spanGaps(self, val):
    self._attrs["spanGaps"] = val


class DataSetBubble(DataAttrs):
  @property
  def backgroundColor(self):
    """
    https://www.chartjs.org/docs/latest/charts/bubble.html
    """
    return self._attrs["backgroundColor"]

  @backgroundColor.setter
  def backgroundColor(self, val):
    self._attrs["backgroundColor"] = val

  @property
  def borderColor(self):
    """
    https://www.chartjs.org/docs/latest/charts/bubble.html
    """
    return self._attrs["borderColor"]

  @borderColor.setter
  def borderColor(self, val):
    self._attrs["borderColor"] = val

  @property
  def borderWidth(self):
    """
    https://www.chartjs.org/docs/latest/charts/bubble.html
    """
    return self._attrs["borderWidth"]

  @borderWidth.setter
  def borderWidth(self, val):
    self._attrs["borderWidth"] = val

  @property
  def data(self):
    """
    https://www.chartjs.org/docs/latest/charts/bubble.html
    """
    return self._attrs["data"]

  @data.setter
  def data(self, val):
    self._attrs["data"] = val

  @property
  def fillOpacity(self):
    """
    Convert the hexadecimal color to the corresponding RGB one with the opacity

    https://www.chartjs.org/docs/latest/charts/doughnut.html
    """
    return self._attrs["backgroundColor"]

  @fillOpacity.setter
  def fillOpacity(self, val):
    bgColors = self._attrs["backgroundColor"]
    if isinstance(bgColors, list):
      opColors = []
      for c in bgColors:
        color = Colors.getHexToRgb(c)
        opColors.append("rgba(%s, %s, %s, %s)" % (color[0], color[1], color[2], val))
      self._attrs["backgroundColor"] = opColors
    else:
      color = Colors.getHexToRgb(self._attrs["backgroundColor"])
      self._attrs["backgroundColor"] = "rgba(%s, %s, %s, %s)" % (color[0], color[1], color[2], val)

  @property
  def hoverBackgroundColor(self):
    """
    https://www.chartjs.org/docs/latest/charts/bubble.html
    """
    return self._attrs["hoverBackgroundColor"]

  @hoverBackgroundColor.setter
  def hoverBackgroundColor(self, val):
    self._attrs["hoverBackgroundColor"] = val

  @property
  def hoverBorderColor(self):
    """
    https://www.chartjs.org/docs/latest/charts/bubble.html
    """
    return self._attrs["hoverBorderColor"]

  @hoverBorderColor.setter
  def hoverBorderColor(self, val):
    self._attrs["hoverBorderColor"] = val

  @property
  def hoverBorderWidth(self):
    """
    https://www.chartjs.org/docs/latest/charts/bubble.html
    """
    return self._attrs["hoverBorderWidth"]

  @hoverBorderWidth.setter
  def hoverBorderWidth(self, val):
    self._attrs["hoverBorderWidth"] = val

  @property
  def hoverRadius(self):
    """
    https://www.chartjs.org/docs/latest/charts/bubble.html
    """
    return self._attrs["hoverRadius"]

  @hoverRadius.setter
  def hoverRadius(self, val):
    self._attrs["hoverRadius"] = val

  @property
  def hitRadius(self):
    """
    https://www.chartjs.org/docs/latest/charts/bubble.html
    """
    return self._attrs["hitRadius"]

  @hitRadius.setter
  def hitRadius(self, val):
    self._attrs["hitRadius"] = val

  @property
  def label(self):
    """
    https://www.chartjs.org/docs/latest/charts/bubble.html
    """
    return self._attrs["label"]

  @label.setter
  def label(self, val):
    self._attrs["label"] = val

  @property
  def order(self):
    """
    https://www.chartjs.org/docs/latest/charts/bubble.html
    """
    return self._attrs["order"]

  @order.setter
  def order(self, val):
    self._attrs["order"] = val

  @property
  def pointStyle(self):
    """
    https://www.chartjs.org/docs/latest/charts/bubble.html
    """
    return self._attrs["pointStyle"]

  @pointStyle.setter
  def pointStyle(self, val):
    self._attrs["pointStyle"] = val

  @property
  def rotation(self):
    """
    https://www.chartjs.org/docs/latest/charts/bubble.html
    """
    return self._attrs["rotation"]

  @rotation.setter
  def rotation(self, val):
    self._attrs["rotation"] = val

  @property
  def radius(self):
    """
    https://www.chartjs.org/docs/latest/charts/bubble.html
    """
    return self._attrs["radius"]

  @radius.setter
  def radius(self, val):
    self._attrs["radius"] = val


# class ChartJsType(object):
#
#   def __init__(self, report, type, data):
#     self._report, self._type, self._data = report, type, data
#     self._data_attrs, self._opts_attrs = {}, {}
#
#   def toStr(self):
#     return '{type: "%s", data: %s}' % (self._type, self._data.toStr())
#
#   def backgroundColor(self, colors):
#     """
#     Data attribute to change the background color (the area between the line chart and the x axis)
#
#     :param colors: Array. The list of colors to add to the data definition
#     :return: The ChartJs configuration dictionary for chaining
#     """
#     self._data_attrs["backgroundColor"] = colors
#     return self
#
#   def pointStyle(self, text):
#     """
#
#     Documentation
#
#     :param text:
#     :return:
#     """
#     self._data_attrs["pointStyle"] = text
#     return self
#
#   def borderColor(self, colors):
#     """
#
#     Documentation
#
#     :param colors:
#     :return:
#     """
#     self._data_attrs["backgroundColor"] = colors
#     return self
#
#   def fill(self, flag):
#     """
#
#     Documentation
#
#     :param flag:
#     :return:
#     """
#     self._data_attrs["fill"] = flag
#     return self
#
#   def steppedLine(self, flag):
#     self._data_attrs["steppedLine"] = flag
#     return self
#
#   def customDataAttr(self, name, value):
#     """
#     Add a bespoke attribute to the ChartJs definition
#
#     Example
#
#     Documentation
#
#     :param name: The name of the option
#     :param value: The value of the option
#
#     :return:
#     """
#     self._data_attrs[name] = JsUtils.jsConvertData(value, None)
#     return self
#
#   def maintainAspectRatio(self, flag):
#     """
#
#     Example
#
#     Documentation
#
#     :param flag:
#     :return:
#     """
#     self._opts_attrs["maintainAspectRatio"] = flag
#     return self
#
#   def responsive(self, flag):
#     self._opts_attrs["responsive"] = flag
#     return self
#
#   def scaleShowLabels(self, flag):
#     self._opts_attrs["scaleShowLabels"] = flag
#     return self
#
#   def customOption(self, name, value):
#     """
#     Add a bespoke option to the ChartJs definition
#
#     :param name: The name of the option
#     :param value: The value of the option
#
#     :return:
#     """
#     self._opts_attrs[name] = JsUtils.jsConvertData(value, None)
#     return self
#
#   @property
#   def legend(self):
#     return OptionsLegend(self._report)
#
#   @property
#   def xAxes(self):
#     self._opts_attrs.setdefault("scales", {})["xAxes"] = []
#     return ChartJsOptScale(self._opts_attrs.setdefault("scales", {})["xAxes"])
#
#   @property
#   def yAxes(self):
#     self._opts_attrs.setdefault("scales", {})["yAxes"] = []
#     return ChartJsOptScale(self._opts_attrs.setdefault("scales", {})["yAxes"])
#
#   def build(self, htmlCode, varName):
#     return "var %s = new Chart(document.getElementById('%s'), %s)" % (varName, htmlCode, self)
#
#
# class ChartJsTypeBar(ChartJsType):
#
#   def __init__(self, report, data, type='bar'):
#     super(ChartJsTypeBar, self).__init__(report, type, data)
#     self._data_attrs, self._opts_attrs = {}, {}
#     self._data_attrs.update({"type": JsUtils.jsConvertData(type, None), "data": data})
#
#   def backgroundColor(self, colors='rgba(0, 0, 0, 0.1)'):
#     """
#
#     Documentation
#     https://www.chartjs.org/docs/latest/charts/bar.html
#
#     :param colors:
#     :return:
#     """
#     if not isinstance(colors, list):
#       colors = [colors]
#     self._data_attrs["backgroundColor"] = colors
#     return self
#
#   def borderColor(self, colors='rgba(0, 0, 0, 0.1)'):
#     """
#
#     Documentation
#     https://www.chartjs.org/docs/latest/charts/bar.html
#
#     :param colors:
#     :return:
#     """
#     if not isinstance(colors, list):
#       colors = [colors]
#     self._data_attrs["borderColor"] = colors
#     return self
#
#   def borderSkipped(self, text='bottom'):
#     """
#     This setting is used to avoid drawing the bar stroke at the base of the fill.
#     In general, this does not need to be changed except when creating chart types that derive from a bar chart.
#
#     Documentation
#     https://www.chartjs.org/docs/latest/charts/bar.html#borderskipped
#
#     :param text: A value among ['bottom', 'left', 'top', 'right', False]
#
#     :return:
#     """
#     skipped_pos = ['bottom', 'left', 'top', 'right', False]
#     if not text in skipped_pos:
#       raise Exception("text value should be in %s" % skipped_pos)
#
#     self._data_attrs["borderSkipped"] = JsUtils.jsConvertData(text, None)
#     return self
#
#   def borderWidth(self, n=0):
#     """
#     If this value is a number, it is applied to all sides of the rectangle (left, top, right, bottom), except borderSkipped.
#     If this value is an object, the left property defines the left border width.
#     Similarly the right, top and bottom properties can also be specified. Omitted borders and borderSkipped are skipped.
#
#     Documentation
#     https://www.chartjs.org/docs/latest/charts/bar.html#borderwidth
#
#     :param n:
#     :return:
#     """
#     self._data_attrs["borderWidth"] = n
#     return self
#
#   def hoverBackgroundColor(self, colors):
#     pass
#
#   def hoverBorderColor(self, colors):
#     pass
#
#   def hoverBorderWidth(self, n):
#     pass
#
#   def label(self, text):
#     """
#     The label for the dataset which appears in the legend and tooltips.
#
#     :param text:
#     :return:
#     """
#     self._data_attrs["label"] = JsUtils.jsConvertData(text, None)
#     return self
#
#   def xAxisID(self, axisId):
#     """
#     The ID of the x axis to plot this dataset on.
#
#     Documentation
#     https://www.chartjs.org/docs/latest/charts/bar.html#general
#
#     :param axisId: The ID of the x axis to plot this dataset on.
#
#     """
#
#   def yAxisID(self, axisId):
#     pass
#
#   @property
#   def scales(self):
#     """
#
#     :rtype: ChartJsOptScaleBar
#     :return:
#     """
#     if not "scales" in self._data_attrs:
#       self._data_attrs["scales"] = ChartJsOptScaleBar(self._report)
#     return self._data_attrs["scales"]
#
#   def __str__(self):
#     return "{%s}" % ", ".join(["%s: %s" % (k, v) for k, v in self._data_attrs.items()])
#
#
# class ChartJsTypeRadar(ChartJsType):
#   def __init__(self, report, data, type='radar'):
#     super(ChartJsTypeRadar, self).__init__(report, type, data)
#     self._datasets, self.__options, self.__config = [], None, None
#
#   def dataset(self, i=None):
#     if i is None:
#       return self._datasets[-1]
#
#     return self._datasets[i]
#
#   def add_dataset(self, data):
#     """
#
#     :param data:
#     :return:
#     """
#     data = DataSetRadar(self._report, attrs={"data": data})
#     self._datasets.append(data)
#     return data
#
#   @property
#   def config(self):
#     """
#
#     :rtype: Config
#     :return:
#     """
#     if self.__config is None:
#       self.__config = Config(self._report)
#     return self.__config
#
#   @property
#   def options(self):
#     """
#
#     :rtype: Options
#     :return:
#     """
#     if self._options is None:
#       self._options = Options(self._report)
#     return self._options
#
#   def toStr(self):
#     print(self.config.toStr())
#     print("{%s}" % ", ".join([d.toStr() for d in self._datasets]))
#

# class ChartJsTypePie(ChartJsType):
#   def __init__(self, report, data, type='pie'):
#     super(ChartJsTypePie, self).__init__(report, type, data)
#     self._datasets, self.__options, self.__config = [], None, None
#
#   def dataset(self, i=None):
#     if i is None:
#       return self._datasets[-1]
#
#     return self._datasets[i]
#
#   def add_dataset(self, data):
#     """
#
#     :param data:
#     """
#     data = DataSetPie(self._report, attrs={"data": data})
#     self._datasets.append(data)
#     return data
#
#   def toStr(self):
#     return "{type: '%s', data: {labels: %s, datasets: [%s]}}" % (self._type, self.labels, ", ".join([d.toStr() for d in self._datasets]))
#


#
# if __name__ == '__main__':
#   # chart_bar = ChartJsTypeBar(None, []).label("test")
#   # chart_bar.scales.barPercentage(0.4).barThickness(3)
#   # chart_bar.scales.gridLines.circular(True).color(["red"]).drawTicks()
#   # print(chart_bar.build("A", "toto"))
#   chart = ChartJsTypeRadar(None, [])
#   #chart.config.duration = 40
#   dataset = chart.add_dataset([])
#   dataset.label = "test"
#   dataset.borderColor = 'red'
#
#   print(chart.toStr())
#
#   data = '''
# backgroundColor	The line fill color.
# borderCapStyle	Cap style of the line. See MDN.
# borderColor	The line color.
# borderDash	Length and spacing of dashes. See MDN.
# borderDashOffset	Offset for line dashes. See MDN.
# borderJoinStyle	Line joint style. See MDN.
# borderWidth	The line width (in pixels).
# clip	How to clip relative to chartArea. Positive value allows overflow, negative value clips that many pixels inside chartArea. 0 = clip at chartArea. Clipping can also be configured per side: clip: {left: 5, top: false, right: -2, bottom: 0}
# fill	How to fill the area under the line. See area charts.
# lineTension	Bezier curve tension of the line. Set to 0 to draw straightlines. This option is ignored if monotone cubic interpolation is used.
# showLine	If false, the line is not drawn for this dataset.
# spanGaps	If true, lines will be drawn between points with no or null data. If false, points with NaN data will create a break in the line.
# '''
#   for r in data.split("\n"):
#     row = r.strip().split()
#     if row:
#       data = {"attr": row[0], 'web': "https://www.chartjs.org/docs/latest/charts/line.html"}
#       print('''
#     @property
#     def %(attr)s(self):
#       """
#       %(web)s
#       """
#       return self._attrs["%(attr)s"]
#
#     @%(attr)s.setter
#     def %(attr)s(self, val):
#       self._attrs["%(attr)s"] = val
#       ''' % data)



from epyk.core.html.options import Options


class OptionsSortable(Options):

  @property
  def group(self):
    """
    Description:
    ------------

    Related Pages:

			https://github.com/SortableJS/Sortable
    """
    return self.get()

  @group.setter
  def group(self, name):
    self.set(name)

  @property
  def sort(self):
    """
    Description:
    ------------
    sorting inside list

    Related Pages:

			https://github.com/SortableJS/Sortable
    """
    return self.get(True)

  @sort.setter
  def sort(self, bool):
    self.set(bool)

  @property
  def delay(self):
    """
    Description:
    ------------
    Time in milliseconds to define when the sorting should start

    Related Pages:

			https://github.com/SortableJS/Sortable
    """
    return self.get(0)

  @delay.setter
  def delay(self, int):
    self.set(int)

  @property
  def delayOnTouchOnly(self):
    """
    Description:
    ------------
    Only delay if user is using touch

    Related Pages:

			https://github.com/SortableJS/Sortable
    """
    return self.get(False)

  @delayOnTouchOnly.setter
  def delayOnTouchOnly(self, bool):
    self.set(bool)

  @property
  def touchStartThreshold(self):
    """
    Description:
    ------------
    px, how many pixels the point should move before cancelling a delayed drag event

    Related Pages:

			https://github.com/SortableJS/Sortable
    """
    return self.get(0)

  @touchStartThreshold.setter
  def touchStartThreshold(self, num):
    self.set(num)

  @property
  def disabled(self):
    """
    Description:
    ------------
    Disables the sortable if set to true.

    Related Pages:

			https://github.com/SortableJS/Sortable
    """
    return self.get(False)

  @disabled.setter
  def disabled(self, bool):
    self.set(bool)

  @property
  def store(self):
    """
    Description:
    ------------
    Disables the sortable if set to true.

    Related Pages:

			https://github.com/SortableJS/Sortable
    """
    return self.get(None)

  @store.setter
  def store(self, bool):
    self.set(bool)

  @property
  def animation(self):
    """
    Description:
    ------------
    ms, animation speed moving items when sorting, `0` — without animation

    Related Pages:

			https://github.com/SortableJS/Sortable
    """
    return self.get(150)

  @animation.setter
  def animation(self, bool):
    self.set(bool)

  @property
  def easing(self):
    """
    Description:
    ------------
    Easing for animation. Defaults to null. See https://easings.net/ for examples.

    Related Pages:

			https://github.com/SortableJS/Sortable
    """
    return self.get("cubic-bezier(1, 0, 0, 1)")

  @easing.setter
  def easing(self, val):
    self.set(val)

  @property
  def handle(self):
    """
    Description:
    ------------
    Drag handle selector within list items

    Related Pages:

			https://github.com/SortableJS/Sortable
    """
    return self.get(".my-handle")

  @handle.setter
  def handle(self, val):
    self.set(val)

  @property
  def filter(self):
    """
    Description:
    ------------
    Selectors that do not lead to dragging (String or Function)

    Related Pages:

			https://github.com/SortableJS/Sortable
    """
    return self.get(".ignore-elements")

  @filter.setter
  def filter(self, val):
    self.set(val)

  @property
  def preventOnFilter(self):
    """
    Description:
    ------------
    Call `event.preventDefault()` when triggered `filter`

    Related Pages:

			https://github.com/SortableJS/Sortable
    """
    return self.get(True)

  @preventOnFilter.setter
  def preventOnFilter(self, bool):
    self.set(bool)

  @property
  def draggable(self):
    """
    Description:
    ------------
    Specifies which items inside the element should be draggable

    Related Pages:

			https://github.com/SortableJS/Sortable
    """
    return self.get(".item")

  @draggable.setter
  def draggable(self, val):
    self.set(val)

  @property
  def dataIdAttr(self):
    """
    Description:
    ------------
    Specifies which items inside the element should be draggable

    Related Pages:

			https://github.com/SortableJS/Sortable
    """
    return self.get("data-id")

  @dataIdAttr.setter
  def dataIdAttr(self, val):
    self.set(val)

  @property
  def ghostClass(self):
    """
    Description:
    ------------
    Class name for the drop placeholder

    Related Pages:

			https://github.com/SortableJS/Sortable
    """
    return self.get("sortable-ghost")

  @ghostClass.setter
  def ghostClass(self, val):
    self.set(val)

  @property
  def chosenClass(self):
    """
    Description:
    ------------
    Class name for the chosen item

    Related Pages:

			https://github.com/SortableJS/Sortable
    """
    return self.get("sortable-chosen")

  @chosenClass.setter
  def chosenClass(self, val):
    self.set(val)

  @property
  def dragClass(self):
    """
    Description:
    ------------
    Class name for the dragging item

    Related Pages:

			https://github.com/SortableJS/Sortable
    """
    return self.get("sortable-drag")

  @dragClass.setter
  def dragClass(self, val):
    self.set(val)

  @property
  def swapThreshold(self):
    """
    Description:
    ------------
    Threshold of the swap zone

    Related Pages:

			https://github.com/SortableJS/Sortable
    """
    return self.get(1)

  @swapThreshold.setter
  def swapThreshold(self, val):
    self.set(val)

  @property
  def invertSwap(self):
    """
    Description:
    ------------
    Will always use inverted swap zone if set to true

    Related Pages:

			https://github.com/SortableJS/Sortable
    """
    return self.get(False)

  @invertSwap.setter
  def invertSwap(self, val):
    self.set(val)

  @property
  def invertedSwapThreshold(self):
    """
    Description:
    ------------
    Threshold of the inverted swap zone (will be set to swapThreshold value by default)

    Related Pages:

			https://github.com/SortableJS/Sortable
    """
    return self.get(1)

  @invertedSwapThreshold.setter
  def invertedSwapThreshold(self, val):
    self.set(val)

  @property
  def direction(self):
    """
    Description:
    ------------
    Direction that the Sortable should sort in. Can be set to 'vertical', 'horizontal', or a function, which will be called whenever a target is dragged over.
    Must return 'vertical' or 'horizontal'.

    Related Pages:

			https://github.com/SortableJS/Sortable
    """
    return self.get('horizontal')

  @direction.setter
  def direction(self, val):
    self.set(val)

  @property
  def forceFallback(self):
    """
    Description:
    ------------
    ignore the HTML5 DnD behaviour and force the fallback to kick in

    Related Pages:

			https://github.com/SortableJS/Sortable
    """
    return self.get(False)

  @forceFallback.setter
  def forceFallback(self, val):
    self.set(val)

  @property
  def fallbackClass(self):
    """
    Description:
    ------------
    Class name for the cloned DOM Element when using forceFallback

    Related Pages:

			https://github.com/SortableJS/Sortable
    """
    return self.get("sortable-fallback")

  @fallbackClass.setter
  def fallbackClass(self, val):
    self.set(val)

  @property
  def fallbackOnBody(self):
    """
    Description:
    ------------
    Appends the cloned DOM Element into the Document's Body

    Related Pages:

			https://github.com/SortableJS/Sortable
    """
    return self.get(False)

  @fallbackOnBody.setter
  def fallbackOnBody(self, val):
    self.set(val)

  @property
  def fallbackTolerance(self):
    """
    Description:
    ------------
    Specify in pixels how far the mouse should move before it's considered as a drag.

    Related Pages:

			https://github.com/SortableJS/Sortable
    """
    return self.get(0)

  @fallbackTolerance.setter
  def fallbackTolerance(self, val):
    self.set(val)

  @property
  def dragoverBubble(self):
    """
    Description:
    ------------
    Specify in pixels how far the mouse should move before it's considered as a drag.

    Related Pages:

			https://github.com/SortableJS/Sortable
    """
    return self.get(False)

  @dragoverBubble.setter
  def dragoverBubble(self, bool):
    self.set(bool)

  @property
  def removeCloneOnHide(self):
    """
    Description:
    ------------
    Remove the clone element when it is not showing, rather than just hiding it

    Related Pages:

			https://github.com/SortableJS/Sortable
    """
    return self.get(True)

  @removeCloneOnHide.setter
  def removeCloneOnHide(self, bool):
    self.set(bool)

  @property
  def emptyInsertThreshold(self):
    """
    Description:
    ------------
    px, distance mouse must be from empty sortable to insert drag element into it

    Related Pages:

			https://github.com/SortableJS/Sortable
    """
    return self.get(5)

  @emptyInsertThreshold.setter
  def emptyInsertThreshold(self, bool):
    self.set(bool)





from epyk.core.html import tables as html_tables


class Pivottable(object):

  def __init__(self, context):
    self.parent = context

  def pivot(self, recordSet=None, rows=None, cols=None, width=(100, '%'), height=(None, 'px'), htmlCode=None,
            helper=None, options=None, profile=False):
    """
    Description:
    -----------
    Create a HTML Pivot table

    Usage::

      Related Pages:

      https://pivottable.js.org/examples/
    https://react-pivottable.js.org/
    https://jsfiddle.net/nicolaskruchten/w86bgq9o/
    """
    table = html_tables.HtmlTablePivot.PivotTable(self.parent.context.rptObj, recordSet, rows, cols, width, height, htmlCode,
                                                  helper, options, profile)
    return table

  def ui(self, recordSet=None, rows=None, cols=None, width=(100, '%'), height=(None, 'px'), htmlCode=None,
            helper=None, options=None, profile=False):
    """
    Description:
    -----------
    Create a HTML Pivot table

    Usage::

      Related Pages:

      https://pivottable.js.org/examples/
    https://react-pivottable.js.org/
    https://jsfiddle.net/nicolaskruchten/w86bgq9o/
    """
    table = html_tables.HtmlTablePivot.PivotUITable(self.parent.context.rptObj, recordSet, rows, cols, width, height, htmlCode,
                                                  helper, options, profile)
    return table

  def sub_total(self, recordSet=None, rows=None, cols=None, width=(100, '%'), height=(None, 'px'), htmlCode=None,
                helper=None, options=None, profile=False):
    """
    Description:
    -----------
    Create a HTML Pivot table

    Usage::

      Related Pages:

      https://pivottable.js.org/examples/
    https://react-pivottable.js.org/
    https://jsfiddle.net/nicolaskruchten/w86bgq9o/
    """
    table = html_tables.HtmlTablePivot.PivotUITable(self.parent.context.rptObj, recordSet, rows, cols, width, height, htmlCode,
                                                  helper, options, profile)
    table.sub_total()
    return table

  def heatmap(self, recordSet=None, rows=None, cols=None, values=None, width=(100, '%'), height=(None, 'px'),
              htmlCode=None, helper=None, options=None, profile=False):
    """
    Description:
    -----------
    Create a HTML Pivot table

    Usage::

    Related Pages:

      https://pivottable.js.org/examples/
    https://react-pivottable.js.org/
    https://jsfiddle.net/nicolaskruchten/w86bgq9o/
    """
    table = html_tables.HtmlTablePivot.PivotTable(self.parent.context.rptObj, recordSet, rows, cols, width, height, htmlCode,
                                                  helper, options, profile)
    table.renderers.heatmap()
    if values is not None:
      table.aggregators.sumOverSum(values)
    return table

  def c3(self, recordSet=None, rows=None, cols=None, width=(100, '%'), height=(None, 'px'), htmlCode=None,
            helper=None, options=None, profile=False):
    """
    Description:
    -----------
    Create a HTML Pivot table

    Usage::

      Related Pages:

      https://pivottable.js.org/examples/
    https://react-pivottable.js.org/
    https://jsfiddle.net/nicolaskruchten/w86bgq9o/
    """
    table = html_tables.HtmlTablePivot.PivotUITable(self.parent.context.rptObj, recordSet, rows, cols, width, height, htmlCode,
                                                  helper, options, profile)
    table.renderers.c3.bar()
    return table

  def d3(self, recordSet=None, rows=None, cols=None, width=(100, '%'), height=(None, 'px'), htmlCode=None,
            helper=None, options=None, profile=False):
    """
    Description:
    -----------
    Create a HTML Pivot table

    Usage::

      Related Pages:

      https://pivottable.js.org/examples/
    https://react-pivottable.js.org/
    https://jsfiddle.net/nicolaskruchten/w86bgq9o/
    """
    table = html_tables.HtmlTablePivot.PivotUITable(self.parent.context.rptObj, recordSet, rows, cols, width, height, htmlCode,
                                                  helper, options, profile)
    table.renderers.treemap()
    return table

  def plotly(self, recordSet=None, rows=None, cols=None, width=(100, '%'), height=(None, 'px'), htmlCode=None,
            helper=None, options=None, profile=False):
    """
    Description:
    -----------
    Create a HTML Pivot table

    Usage::

      Related Pages:

      https://pivottable.js.org/examples/
    https://react-pivottable.js.org/
    https://jsfiddle.net/nicolaskruchten/w86bgq9o/
    """
    table = html_tables.HtmlTablePivot.PivotUITable(self.parent.context.rptObj, recordSet, rows, cols, width, height, htmlCode,
                                                    helper, options, profile)
    table.renderers.plotly.bar()
    return table




from epyk.core.html import Html
from epyk.core.html.options import OptSliders
from epyk.core.html.options import OptList

from epyk.core.html.entities import EntHtml4

from epyk.core.js import JsUtils
from epyk.core.js.html import JsHtmlJqueryUI
from epyk.core.js.html import JsHtmlList
from epyk.core.js.packages import JsQuery
from epyk.core.js.packages import JsQueryUi

# The list of CSS classes
from epyk.core.css.styles import GrpClsJqueryUI


class ProgressBar(Html.Html):
  requirements = ('jqueryui', )
  name = 'Progress Bar'

  def __init__(self, report, number, total, width, height, attrs, helper, options, htmlCode, profile):
    options['max'] = total
    super(ProgressBar, self).__init__(report, number, htmlCode=htmlCode, css_attrs={"width": width, "height": height, 'box-sizing': 'border-box'}, profile=profile)
    self.add_helper(helper)
    self.__options = OptSliders.OptionsProgBar(self, options)
    self.options.background = self._report.theme.success[1]

  @property
  def options(self):
    """
    Description:
    -----------
    The progress bar is designed to display the current percent complete for a process.
    The bar is coded to be flexibly sized through CSS and will scale to fit inside its parent container by default.

    Related Pages:

			https://api.jqueryui.com/progressbar

    :rtype: OptSliders.OptionsProgBar
    """
    return self.__options

  @property
  def _js__builder__(self):
    return '''
      options.value = parseFloat(data);
      %(jqId)s.progressbar(options).find('div').attr("data-toggle", "tooltip").attr("title", ""+ (parseFloat(data) / options.max * 100).toFixed(2) +"%% ("+ parseFloat(data) +" / "+ options.max +")").css(options.css);
      ''' % {"jqId": JsQuery.decorate_var("htmlObj", convert_var=False)}

  @property
  def js(self):
    """
    Description:
    -----------
    Javascript Functions

    Return all the Javascript functions defined for an HTML Component.
    Those functions will use plain javascript by default.

    Related Pages:

			https://api.jqueryui.com/progressbar

    Attributes:
    ----------
    :return: A Javascript Dom object

    :rtype: JsQueryUi.ProgressBar
    """
    if self._js is None:
      self._js = JsQueryUi.ProgressBar(self, report=self._report)
    return self._js

  @property
  def dom(self):
    """
    Description:
    ------------
    Javascript Functions

    Return all the Javascript functions defined for an HTML Component.
    Those functions will use plain javascript by default.

    :return: A Javascript Dom object

    :rtype: JsHtml.JsHtmlProgressBar
    """
    if self._dom is None:
      self._dom = JsHtmlJqueryUI.JsHtmlProgressBar(self, report=self._report)
    return self._dom

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    return '<div %s></div>%s' % (self.get_attrs(pyClassNames=self.style.get_classes()), self.helper)


class Menu(Html.Html):
  requirements = ('jqueryui', )
  name = 'Menu'

  def __init__(self, report, records, width, height, attrs, helper, options, htmlCode, profile):
    super(Menu, self).__init__(report, records, css_attrs={"width": width, "height": height}, profile=profile)
    self.add_helper(helper)
    self.__options = OptSliders.OptionsMenu(self, options)
    self.css({"display": 'block', 'position': 'relative'})

  @property
  def options(self):
    """
    Description:
    -----------

    Related Pages:

			https://api.jqueryui.com/menu

    :rtype: OptSliders.OptionsMenu
    """
    return self.__options

  @property
  def _js__builder__(self):
    return '''
          var jqHtmlObj = jQuery(htmlObj); if (options.clearDropDown) {jqHtmlObj.empty()};
          data.forEach(function(rec){
            if (rec.items != undefined) {
              var li = $('<li></li>'); var div = $('<div>'+ rec.value +'</div>').css({"width": '150px'});
              li.append(div); var ul = $('<ul aria-hidden="true"></ul>'); options.clearDropDown = false;
              %(pyCls)s(ul, rec.items, options); li.append(ul); jqHtmlObj.append(li);
            } else {
              var div = $('<div>'+ rec.value +'</div>').css({"width": '150px'}); var li = $('<li></li>');
              li.append(div); jqHtmlObj.append(li)};
          }); jqHtmlObj.menu(options)''' % {"pyCls": self.__class__.__name__}

  @property
  def js(self):
    """
    Description:
    -----------
    Javascript Functions

    Related Pages:

			https://api.jqueryui.com/menu

    Attributes:
    ----------
    :return: A Javascript Dom object

    :rtype: JsQueryUi.Menu
    """
    if self._js is None:
      self._js = JsQueryUi.Menu(self, report=self._report)
    return self._js

  @property
  def dom(self):
    """
    Description:
    -----------
    Javascript Functions

    Return all the Javascript functions defined for an HTML Component.
    Those functions will use plain javascript by default.

    :return: A Javascript Dom object
    :rtype: JsHtml.JsHtmlProgressBar
    """
    if self._dom is None:
      self._dom = JsHtmlJqueryUI.JsHtmlProgressBar(self, report=self._report)
    return self._dom

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    return '<ul %s></ul>%s' % (self.get_attrs(pyClassNames=self.style.get_classes()), self.helper)


class Dialog(Html.Html):
  requirements = ('jqueryui', )
  name = 'Menu'

  def __init__(self, report, text, width, height, attrs, helper, options, htmlCode, profile):
    super(Dialog, self).__init__(report, text, css_attrs={"width": width, "height": height}, profile=profile)
    self.add_helper(helper)
    self.__options = OptSliders.OptionDialog(self, options)

  @property
  def options(self):
    """
    Description:
    -----------
    Open content in an interactive overlay.

    Related Pages:

			https://jqueryui.com/dialog/

    :rtype: OptSliders.OptionDialog
    """
    return self.__options

  @property
  def _js__builder__(self):
    return "jQuery(htmlObj).empty(); jQuery(htmlObj).append('<p>'+ data +'</p>'); jQuery(htmlObj).dialog(options)"

  @property
  def js(self):
    """
    Description:
    -----------
    Open content in an interactive overlay.

    Related Pages:

			https://jqueryui.com/dialog/

    Attributes:
    ----------
    :return: A Javascript Dom object

    :rtype: JsQueryUi.Dialog
    """
    if self._js is None:
      self._js = JsQueryUi.Dialog(self, report=self._report)
    return self._js

  @property
  def dom(self):
    """
    Javascript Functions

    Return all the Javascript functions defined for an HTML Component.
    Those functions will use plain javascript by default.

    :return: A Javascript Dom object
    :rtype: JsHtml.JsHtmlProgressBar
    """
    if self._dom is None:
      self._dom = JsHtmlJqueryUI.JsHtmlProgressBar(self, report=self._report)
    return self._dom

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    return '<div %s>%s</div>' % (self.get_attrs(pyClassNames=self.style.get_classes()), self.helper)


class Slider(Html.Html):
  requirements = ('jqueryui', )
  name = 'Slider'

  def __init__(self, report, number, min, max, width, height, attrs, helper, options, htmlCode, profile):
    options.update({'max': max, 'min': min})
    super(Slider, self).__init__(report, number, htmlCode=htmlCode, css_attrs={"width": width, "height": height}, profile=profile)
    self._jsStyles = {'css': {"background": self._report.theme.success[0]}}
    self.__options = OptSliders.OptionsSlider(self, options)
    self.style.css.padding = "0 10px"
    self.style.css.margin = "15px 0"

  @property
  def options(self):
    """

    :rtype: OptSliders.OptionsSlider
    """
    return self.__options

  @property
  def style(self):
    """
    Description:
    ------------
    Property to the CSS Style of the component

    :rtype: GrpClsJqueryUI.ClassSlider
    """
    if self._styleObj is None:
      self._styleObj = GrpClsJqueryUI.ClassSlider(self)
    return self._styleObj

  @property
  def js(self):
    """
    Description:
    -----------
    Javascript Functions

    Return all the Javascript functions defined for an HTML Component.
    Those functions will use plain javascript by default.

    Related Pages:

			https://api.jqueryui.com/slider

    Attributes:
    ----------
    :return: A Javascript Dom object

    :rtype: JsQueryUi.Slider
    """
    if self._js is None:
      self._js = JsQueryUi.Slider(self, report=self._report)
    return self._js

  def change(self, jsFnc, profile=None):
    """
    Description:
    -----------
    Triggered after the user slides a handle, if the value has changed; or if the value is changed programmatically via the value method.

    Related Pages:

			https://api.jqueryui.com/slider/#event-change

    Attributes:
    ----------
    :param jsFnc:
    """
    if not isinstance(jsFnc, list):
      jsFnc = [jsFnc]
    self._jsStyles["change"] = "function(event, ui){ %s }" % JsUtils.jsConvertFncs(jsFnc, toStr=True)
    return self

  def start(self, jsFnc, profile=None):
    """
    Description:
    -----------
    Triggered when the user starts sliding.

    Related Pages:

			https://api.jqueryui.com/slider/#event-start

    Attributes:
    ----------
    :param jsFnc:
    """
    if not isinstance(jsFnc, list):
      jsFnc = [jsFnc]
    self._jsStyles["start"] = "function(event, ui){ %s }" % JsUtils.jsConvertFncs(jsFnc, toStr=True)
    return self

  def slide(self, jsFnc, profile=None):
    """
    Description:
    -----------
    Triggered when the user starts sliding.

    Related Pages:

			https://api.jqueryui.com/slider/#event-slide

    Attributes:
    ----------
    :param jsFnc:
    """
    if not isinstance(jsFnc, list):
      jsFnc = [jsFnc]
    self._jsStyles["slide"] = "function(event, ui){ %s }" % JsUtils.jsConvertFncs(jsFnc, toStr=True)
    return self

  def stop(self, jsFnc, profile=None):
    """
    Description:
    -----------
    Triggered after the user slides a handle.

    Related Pages:

			https://api.jqueryui.com/slider/#event-stop

    Attributes:
    ----------
    :param jsFnc:
    """
    if not isinstance(jsFnc, list):
      jsFnc = [jsFnc]
    self._jsStyles["stop"] = "function(event, ui){ %s }" % JsUtils.jsConvertFncs(jsFnc, toStr=True)
    return self

  @property
  def dom(self):
    """
    Description:
    -----------
    The Javascript Dom object

    :rtype: JsHtmlJqueryUI.JsHtmlSlider
    """
    if self._dom is None:
      self._dom = JsHtmlJqueryUI.JsHtmlSlider(self, report=self._report)
    return self._dom

  @property
  def _js__builder__(self):
    return ''' options.value = data; %(jqId)s.slider(options).css(options.css)
      ''' % {"jqId": JsQuery.decorate_var("htmlObj", convert_var=False)}

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    return '''
      <div %(strAttr)s>
        <div style="width:100%%;height:20px">
          <span style="float:left;display:inline-block">%(min)s</span>
          <span style="float:right;display:inline-block">%(max)s</span>
        </div>
        <div id="%(htmlCode)s"></div>
      </div>%(helper)s''' % {"strAttr": self.get_attrs(withId=False), "min": self.options.min, "htmlCode": self.htmlCode,
                             "max": self.options.max, "helper": self.helper}


class Range(Slider):

  @property
  def _js__builder__(self):
    return '''options.values = [Math.min(...data), Math.max(...data)]; %(jqId)s.slider(options).css(options.css)
        ''' % {"jqId": JsQuery.decorate_var("htmlObj", convert_var=False)}


class SliderDate(Slider):

  def __init__(self, report, number, min, max, width, height, attrs, helper, options, htmlCode, profile):
    super(SliderDate, self).__init__(report, number, min, max, width, height, attrs, helper, options, htmlCode, profile)
    self.options.min = min
    self.options.max = max
    self.options.step = 86400

  @property
  def _js__builder__(self):
    return '''
      const minDt = new Date(options.min).getTime() / 1000;
      const maxDt = new Date(options.max).getTime() / 1000;
      
      options.min = minDt; options.max = maxDt;
      options.value = new Date(data).getTime() / 1000;
      %(jqId)s.slider(options).css(options.css)
      ''' % {"jqId": JsQuery.decorate_var("jQuery(htmlObj)", convert_var=False)}

  @property
  def dom(self):
    """
    Description:
    -----------
    The Javascript Dom object

    :rtype: JsHtmlJqueryUI.JsHtmlSlider
    """
    if self._dom is None:
      self._dom = JsHtmlJqueryUI.JsHtmlSliderDate(self, report=self._report)
    return self._dom


class SliderDates(SliderDate):

  @property
  def _js__builder__(self):
    return '''
      const minDt = new Date(options.min).getTime() / 1000;
      const maxDt = new Date(options.max).getTime() / 1000;

      options.min = minDt; options.max = maxDt;
      options.values = [new Date(data[0]).getTime() / 1000, new Date(data[1]).getTime() / 1000];
      %(jqId)s.slider(options).css(options.css)
      ''' % {"jqId": JsQuery.decorate_var("jQuery(htmlObj)", convert_var=False)}

  @property
  def dom(self):
    """
    Description:
    -----------
    The Javascript Dom object

    :rtype: JsHtmlJqueryUI.JsHtmlSlider
    """
    if self._dom is None:
      self._dom = JsHtmlJqueryUI.JsHtmlSliderDates(self, report=self._report)
    return self._dom


class SkillBar(Html.Html):
  name = 'Skill Bars'

  def __init__(self, report, data, y_column, x_axis, title, width, height, htmlCode, profile):
    super(SkillBar, self).__init__(report, "", css_attrs={"width": width, "height": height}, htmlCode=htmlCode, profile=profile)
    self.add_title(title, options={'content_table': False})
    self.innerPyHTML = report.ui.layouts.table(options={"header": False}) # data, y_column, x_axis)
    self.innerPyHTML.options.managed = False
    for rec in data:
      value = report.ui.div(EntHtml4.NO_BREAK_SPACE).css({"width": '%spx' % rec[y_column], 'margin-left': "2px",  "background": report.theme.success[0]})
      value.options.managed = False
      self.innerPyHTML += [rec[x_axis], value]
      self.innerPyHTML[-1][1].attr["align"] = 'left'
    self.innerPyHTML.style.clear()
    self.css({"margin": '5px 0'})

  @property
  def _js__builder__(self):
    return '''
      htmlObj.empty();
      data.forEach(function(rec, i){
        if (options.colTooltip != undefined) {var tooltip = 'title="'+ rec[options.colTooltip] +'"'} else {var tooltip = ''};
        if (options.colUrl != undefined) {var content = '<a href="'+ rec[options.colUrl] +'" style="color:white">'+ valPerc.toFixed(2) +'%</a>'} 
        else {var content = rec[options.val].toFixed( 2 ) +"%"};
        htmlObj.append('<tr '+ tooltip +'><td style="width:100px;"><p style="margin:2px;text-align:center;word-wrap:break-word;cursor:pointer">'+ rec[options.label] +'</p></td><td style="width:100%"><div style="margin:2px;display:block;height:100%;padding-bottom:5px;padding:2px 0 2px 5px;width:'+ parseInt( rec[options.val] ) +'%;background-color:'+ jsStyles.color +';color:'+ options.fontColor +'">' + content + '</div></td></tr>');
        htmlObj.find('tr').tooltip()})
      '''

  def __str__(self):
    return '<div %s>%s</div>' % (self.get_attrs(pyClassNames=self.style.get_classes()), self.content)


class OptionsBar(Html.Html):
  requirements = ('font-awesome', )
  name = 'Options'

  def __init__(self, report, recordset, width, height, color, options):
    super(OptionsBar, self).__init__(report, [], css_attrs={"width": width, 'height': height})
    self.__options = OptSliders.OptionBar(self, options)
    self.css({'padding': '0', 'display': 'block', 'text-align': 'middle', 'color': color, 'margin-left': '5px',
              'background': self._report.theme.greys[0]})
    for rec in recordset:
      self += rec
    if self.options.draggable:
      self.draggable()

  @property
  def options(self):
    """
    :rtype: OptSliders.OptionBar
    """
    return self.__options

  def __add__(self, icon):
    """ Add items to a container """
    icon = self._report.ui.icon(icon)
    icon.style.css.margin = "5px"
    super(OptionsBar, self).__add__(icon)
    return self

  def draggable(self, options=None):
    self.css({'border-radius': '5px', "border": "1px dotted %s" % self._report.theme.success[1]})
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.dom.jquery_ui.draggable(options).toStr())
    return self

  def __str__(self):
    str_html = "".join([v.html() for v in self.val])
    return '<div %(attrs)s>%(icons)s</div>' % {'attrs': self.get_attrs(pyClassNames=self.style.get_classes()), 'icons': str_html}


class SignIn(Html.Html):
  requirements = ('font-awesome', )
  name = 'SignIn'

  def __init__(self, report, text, size, icon):
    super(SignIn, self).__init__(report, text, css_attrs={"width": size, 'height': size})
    self.size, self.icon = "%s%s" % (size[0]-8, size[1]), icon
    self.css({"text-align": "center", "padding": 0, 'color': self._report.theme.colors[3],
              "margin": 0, "border-radius": "%s%s" % (size[0], size[1]),
              "border": "1px solid %s" % self._report.theme.colors[3], 'cursor': 'pointer'})

  def __str__(self):
    if not hasattr(self._report, 'user') or self._report.user == 'local':
      self.attr["class"].add(self.icon or "fas fa-user-tie")
      self.style.css.font_family = "Font Awesome 5 Free"
      self.style.css.padding = "2px"
      self.style.css.font_size = self.size
      return '<i title="Guest Mode" %(attrs)s></i>' % {'size': self.size, 'attrs': self.get_attrs(pyClassNames=self.style.get_classes())}

    return '''
      <div title="%(user)s" %(attrs)s>
        <p style="font-size:%(size)s;margin-top:-2px">%(letter)s</p>
      </div> ''' % {'size': self.size, 'letter': self._report.user[0].upper(), 'user': self._report.user,
                    'attrs': self.get_attrs(pyClassNames=self.style.get_classes())}


class Filters(Html.Html):
  name = 'Filters'
  requirements = ('font-awesome', )

  def __init__(self, report, items, width, height, htmlCode, helper, options, profile):
    super(Filters, self).__init__(report, items, css_attrs={"width": width, "height": height}, htmlCode=htmlCode, profile=profile)
    self.__options = OptList.OptionsTagItems(self, options)
    self._jsStyles['delete'] = 'function(){this.parentNode.remove()}'
    self.input = self._report.ui.input()
    self.input.style.css.text_align = 'left'
    self.input.style.css.padding = '0 5px'
    self.input.options.managed = False
    self.selections = self._report.ui.div()
    self.selections.options.managed = False
    self.selections.attr["name"] = "panel"
    self.selections.css({'height': '30px', 'padding': '5px 2px'})
    self.add_helper(helper)
    self.__enter_def = False

  @property
  def options(self):
    """
    :rtype: OptList.OptionsTagItems
    """
    return self.__options

  @property
  def _js__builder__(self):
    return '''
      var panel = htmlObj.querySelector('[name=panel]'); panel.innerHTML = '';
      data.forEach(function(val){
        if(typeof val === 'string'){ val = {name: options.category, category: options.category, value: val, disabled: false, fixed: false} }
        else{
          if(val.category === undefined){ if(val.name === undefined) {val.category = options.category} else {val.category = val.name}}
          if(val.name === undefined){ val.name = val.category }};
        chipAdd(panel, val, options)})'''

  def enter(self, jsFncs, profile=False):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param jsFncs:
    :param profile:
    """
    self.__enter_def = True
    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    self.keydown.enter([JsUtils.jsConvertFncs(jsFncs, toStr=True), self.dom.add(self.dom.input)] + jsFncs + [self.input.dom.empty()], profile)
    return self

  def drop(self, jsFncs, preventDefault=True, profile=False):
    """
    Description:
    -----------

    :param jsFncs:
    :param preventDefault:
    :param profile:
    """
    self.style.css.border = "1px dashed black"
    self.tooltip("Drag and drop values here")
    return super(Filters, self).drop(jsFncs, preventDefault, profile)

  def delete(self, jsFncs, profile=False):
    """
    Description:
    -----------

    :param jsFncs:
    :param profile:
    """
    if self.__enter_def:
      raise Exception("delete on chip must be triggered before enter")

    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    self._jsStyles['delete'] = 'function(){ this.parentNode.remove(); %s}' % JsUtils.jsConvertFncs(jsFncs, toStr=True)
    return self # function(event){ this.parentNode.remove()}

  def append(self, value, category=None, name=None, disabled=False, fixed=False):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param value:
    :param category:
    :param disabled:
    :param fixed:
    """
    rec = {"value": value, 'disabled': disabled, 'fixed': fixed, 'category': category, 'name': name}
    if category is None:
      rec['category'] = name or self.options.category
    rec['name'] = name or rec['category']
    self._vals.append(rec)

  @property
  def dom(self):
    """
    Description:
    -----------
    The Javascript Dom object

    :rtype: JsHtmlList.Tags
    """
    if self._dom is None:
      self._dom = JsHtmlList.Tags(self, report=self._report)
    return self._dom

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    #
    constructors = self._report._props.setdefault("js", {}).setdefault("constructors", {})
    constructors['ChipAdd'] = '''function chipAdd(panel, record, options){
        var div = document.createElement("div"); for (var key in options.item_css){ div.style[key] = options.item_css[key]};
        div.setAttribute('data-category', record.category);
        var content = document.createElement("span"); for (var key in options.value_css){ content.style[key] = options.value_css[key]};
        content.setAttribute('name', 'chip_value'); content.innerHTML = record.value; 
        if(options.visible){
          var p = document.createElement("p"); for (var key in options.category_css){ p.style[key] = options.category_css[key]};
          p.innerHTML = record.name; div.appendChild(p)
        };
        div.appendChild(content);
        if(!record.fixed){
          var icon = document.createElement("i"); for (var key in options.icon_css){ icon.style[key] = options.icon_css[key] };
          icon.classList.add('fas'); icon.classList.add('fa-times');  icon.addEventListener('click', options.delete );
          div.appendChild(icon)}
        panel.appendChild(div);
    }'''
    if not self.options.visible:
      self.input.style.css.display = False
    return '''<div %(attrs)s>%(input)s%(selections)s</div>%(helper)s''' % {'attrs': self.get_attrs(pyClassNames=self.style.get_classes()),
          'input': self.input.html(), 'selections': self.selections.html(),  'helper': self.helper}




from epyk.core.js.html import JsHtml
from epyk.core.js.primitives import JsObjects


class TimePicker(JsHtml.JsHtml):

  @property
  def val(self):
    return JsObjects.JsObjects.get(
      "{%s: {value: %s.timepicker('getTime'), timestamp: Date.now(), offset: new Date().getTimezoneOffset()}}" % (
        self.htmlCode, self._src.dom.jquery.varId))

  @property
  def content(self):
    return JsHtml.ContentFormatters(self._report, "%s.timepicker('getTime')" % self._src.dom.jquery.varId)




import collections

from epyk.core.html import Html

from epyk.core.js import expr
from epyk.core.css import Selector

# The list of CSS classes
from epyk.core.css.styles import GrpClsMenu
from epyk.core.html.HtmlList import Li
from epyk.core.html.options import OptList


class HtmlNavBar(Html.Html):
  name = 'Nav Bar'

  def __init__(self, report, components, width, height, options, profile):
    super(HtmlNavBar, self).__init__(report, [], css_attrs={"width": width, "height": height}, profile=profile)
    if components is not None:
      if not isinstance(components, list):
        components = [components]
      for c in components:
        self.__add__(c)

  @property
  def style(self):
    """
    Description:
    -----------
    Property to the CSS Style of the component

    :rtype: GrpClsMenu.ClassNav
    """
    if self._styleObj is None:
      self._styleObj = GrpClsMenu.ClassNav(self)
    return self._styleObj

  def move(self):
    """

    """
    super(HtmlNavBar, self).move()
    self.style.css.position = None
    self._report.body.style.css.padding_top = 0

  def __add__(self, htmlObj):
    """ Add items to the footer """
    htmlObj.options.managed = False # Has to be defined here otherwise it is set to late
    htmlObj.style.css.display = 'inline'
    if htmlObj.css('height') is None:
      htmlObj.style.css.vertical_align = 'middle'
    if htmlObj.css('width') == '100%':
      htmlObj.style.css.width = None
    self.val.append(htmlObj)
    return self

  def add_text(self, text):
    """

    :param text:
    :return:
    """
    val = self._report.ui.text(text)
    self.__add__(val)
    val.style.css.height = "100%"
    val.style.css.vertical_align = 'middle'
    return val

  def __str__(self):
    str_h = "".join([h.html() for h in self.val])
    return "<div %s>%s</div>" % (self.get_attrs(pyClassNames=self.style.get_classes()), str_h)


class HtmlFooter(Html.Html):
  name = 'footer'

  def __init__(self, report, components, width, height, profile):
    super(HtmlFooter, self).__init__(report, [], css_attrs={"width": width, "height": height}, profile=profile)
    self.__col_lst = None
    if components is not None:
      if not isinstance(components, list):
        components = [components]
      for c in components:
        self.__add__(c)

  @property
  def sections(self):
    if not self.__col_lst:
      self.__col_lst = []
    return self.__col_lst

  @sections.setter
  def sections(self, col_lst):
    self.__col_lst = col_lst

  @property
  def style(self):
    """
    Description:
    -----------
    Property to the CSS Style of the component

    :rtype: GrpClsMenu.ClassFooter
    """
    if self._styleObj is None:
      self._styleObj = GrpClsMenu.ClassFooter(self)
    return self._styleObj

  def __add__(self, htmlObj):
    """ Add items to the footer """
    if not hasattr(htmlObj, 'options'):
      htmlObj = self._report.ui.div(htmlObj)
    htmlObj.options.managed = False # Has to be defined here otherwise it is set to late
    self.val.append(htmlObj)
    return self

  def __getitem__(self, i):
    """
    Return the internal column in the row for the given index

    :param i: the column index
    """
    return self.val[i]

  def add_menu(self, context_menu):
    pass

  def __str__(self):
    str_h = "".join([val.html() for val in self.val])
    return "<footer %s>%s</footer>" % (self.get_attrs(pyClassNames=self.style.get_classes()), str_h)


class ContextMenu(Html.Html):
  name = 'Context Menu'
  source = None # The container

  def __init__(self, report, recordSet, width, height, visible, options, profile):
    super(ContextMenu, self).__init__(report, [], css_attrs={"width": width, "height": height}, profile=profile)
    self.__options = OptList.OptionsLi(self, options)
    self.css({'display': 'block' if visible else 'none', 'position': 'absolute', 'z-index': 200,
              'padding': 0, 'margin': 0, 'background-color': self._report.theme.greys[0],
              'border': '1px solid %s' % self._report.theme.success[0], 'border-radius': '2px'})
    self.style.css.shadow_box()
    for rec in recordSet:
      self.__add__(rec)

  @property
  def options(self):
    """

    :rtype: OptList.OptionsLi
    """
    return self.__options

  def add_item(self, value, icon=None):
    """

    :param value:
    :param icon:
    """
    self += {"value": value, 'icon': icon}
    return self

  def add(self, htmlObj):
    """

    :param htmlObj:
    """
    self.__add__(htmlObj)
    return self.val[-1].val

  def __add__(self, htmlObj):
    """

    :param d:
    """
    if not hasattr(htmlObj, 'options'):
      if isinstance(htmlObj, dict):
        if htmlObj['icon'] is not None:
          i = self._report.ui.icon(htmlObj['icon'])
          i.css({'display': 'inline', 'margin-right': '5px'})
          v = self._report.ui.text(htmlObj['value'])
          v.css({'display': 'inline'})
          htmlObj = self._report.ui.div([i, v])
        else:
          htmlObj = self._report.ui.div(htmlObj['value'])
      else:
        htmlObj = self._report.ui.div(htmlObj)
    li_obj = Li(self._report, htmlObj) if not isinstance(htmlObj, Li) else htmlObj
    li_obj.css({"padding": "5px", 'cursor': 'pointer'})
    li_obj.options.managed = False
    htmlObj.options.managed = False
    self.val.append(li_obj)
    return self

  def __getitem__(self, i):
    return self.val[i].val

  def __str__(self):
    #self._report._scroll.add("$('nav[name=context_menus]').hide()")
    # TODO: Add a condition in the init to display the context menu only for some columns or rows when table for example
    if getattr(self, 'source') is None:
      raise Exception("Context Menu should be added to a component with the function contextMenu")

    str_vals = "".join([i.html() for i in self.val]) if self.val is not None else ""
    self.mouse(out_fncs=[self.dom.hide()]) # hide when mouse leave the component
    return '''
      <nav %(attr)s name='context_menus'>
        <ul style='list-style:none;padding:0px;margin:0'>%(val)s</ul>
      </nav>''' % {'attr': self.get_attrs(pyClassNames=self.style.get_classes()), 'val': str_vals}


class PanelsBar(Html.Html):
  name = 'Pabel'

  def __init__(self, report, width, height, options, helper, profile):
    super(PanelsBar, self).__init__(report, None, css_attrs={"width": width, "height": height})
    self.menus = report.ui.div(options={'inline': True})
    self.menus.options.managed = False
    self.panels = report.ui.div()
    self.panels.options.managed = False
    self.menu_mapping = collections.OrderedDict()
    #
    self.panels.style.css.display = False
    self.panels.style.css.position = 'absolute'
    #
    self.style.css.position = 'relative'
    self.__options = options
    if options.get('position', 'top') == 'top':
      self.panels.style.css.padding_bottom = 10
      self.menus.style.css.top = 0
      self.panels.style.css.border_bottom = '1px solid %s' % report.theme.colors[-1]
    else:
      self.style.css.position = 'relative'
      self.panels.style.css.border_top = '1px solid %s' % report.theme.colors[-1]
      self.panels.style.css.bottom = 0
      self.menus.style.css.bottom = 0

    self.menus.style.css.position = 'absolute'
    self.menus.style.css.background = report.theme.colors[-1]
    self.menus.style.css.color = report.theme.colors[0]
    self.menus.style.css.padding = '5px 0'

  def add_panel(self, text, content):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param text:
    :param content:
    """
    content.style.css.padding = "0 5px"
    if not hasattr(text, 'htmlCode'):
      text = self._report.ui.div(text)
    text.style.css.display = 'inline-block'
    text.style.css.width = 'auto'
    text.style.css.cursor = 'pointer'
    text.style.css.padding = '0 5px'
    self.menu_mapping[text] = content
    self.menus += text
    self.panels += content
    return self

  def __str__(self):
    css_menu = {"height": "auto", 'display': 'block', 'margin-top': '30px'} if self.__options['position'] == 'top' else {"height": "auto", 'display': 'block', 'margin-bottom': '30px'}
    for menu_item, panel in self.menu_mapping.items():
      menu_item.click([
        self._report.js.querySelectorAll(Selector.Selector(self.panels).with_child_element("div").excluding(panel)).css({"display": 'none'}),
        #
        expr.if_(self._report.js.querySelector(Selector.Selector(self.panels)).getAttribute("data-panel") == menu_item.htmlCode, [
          self._report.js.querySelector(Selector.Selector(self.panels)).setAttribute("data-panel", ""),
          self._report.js.querySelector(Selector.Selector(self.panels)).css({"display": 'none'})
        ]).else_([
          self._report.js.querySelector(Selector.Selector(self.panels)).setAttribute("data-panel", menu_item.htmlCode),
          self._report.js.querySelector(Selector.Selector(self.panels)).css(css_menu),
          self._report.js.querySelector(Selector.Selector(panel)).css({'display': 'block'})
        ])
      ])

    return "<div %s>%s%s</div>" % (self.get_attrs(pyClassNames=self.style.get_classes()), self.menus.html(), self.panels.html())


class Shortcut(Html.Html):
  name = 'shortcut'

  def __init__(self, report, components, width, height, htmlCode, options, profile):
    super(Shortcut, self).__init__(report, [], htmlCode=htmlCode, css_attrs={"width": width, "height": height}, profile=profile)
    self.logo = None
    self.__options = options
    for component in components:
      if not hasattr(component, 'options'):
        component = report.ui.icons.awesome(component)
      self.__add__(component)
    self.css({"background": report.theme.colors[1], "position": 'fixed', 'overflow': 'hidden', 'z-index': 20})
    self.style.css.padding = 0
    if options['position'] in ['left', 'right']:
      self.css({'text-align': 'center'})
    elif options['position'] == 'top':
      self.css({'top': '0'})
    elif options['position'] == 'bottom':
      self.css({'bottom': '0'})

  @property
  def style(self):
    """
    Description:
    -----------

    :rtype: GrpClsMenu.ClassShortcut
    """
    if self._styleObj is None:
      self._styleObj = GrpClsMenu.ClassShortcut(self)
    return self._styleObj

  def __add__(self, component):
    """ Add items to a container """
    if hasattr(component, 'htmlCode'):
      component.options.managed = False

    if self.__options['position'] in ['left', 'right']:
      component.style.css.width = self.css("width")
      component.style.css.text_align = 'center'
      component.style.css.display = 'block'
      component.style.css.margin_bottom = 5

    if self.__options['position'] == 'top':
      component.style.css.line_height = self.css("height")
      component.style.css.height = self.css("height")
      component.style.css.vertical_align = 'top'
      if component.style.css.position is None:
        component.style.css.position = 'relative'
      component.style.css.top = -3
      component.style.css.display = 'inline-block'

    component.style.css.margin = 'auto'
    self.val.append(component)
    return self

  def add_logo(self, icon, path=None, align="center", width=(32, 'px'), height=(32, 'px')):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param icon:
    :param path:
    :param align:
    :param width:
    :param height:
    """
    self.logo = self._report.ui.img(icon, path=path, align=align, width=width, height=height)
    self.logo.style.css.margin_top = 5
    if self.__options['position'] in ['left', 'right']:
      self.logo.style.css.margin_bottom = 15
      self.logo.style.css.margin_right = 0
    else:
      self.logo.style.css.margin_right = 10
    return self

  def __str__(self):
    if self.logo is None:
      self.logo = self._report.ui.icons.epyk()
    self.logo.options.managed = False
    str_div = "".join([self.logo.html()] + [v.html() if hasattr(v, 'html') else str(v) for v in self.val])
    return "<div %s>%s</div>" % (self.get_attrs(pyClassNames=self.style.get_classes()), str_div)




from epyk.interfaces.geo import CompGeoPlotly
from epyk.interfaces.geo import CompGeoDc
from epyk.interfaces.geo import CompGeoChartJs
from epyk.interfaces.geo import CompGeoGoogle


class Geo(object):
  def __init__(self, context):
    self.context = context

  @property
  def plotly_map(self):
    """
    Description:
    ------------
    Interface for the Plotly library
    """
    return CompGeoPlotly.Plotly(self)

  @property
  def dc_choropleth(self):
    """
    Description:
    ------------
    """
    return CompGeoDc.Dc(self)

  @property
  def chartJs(self):
    """
    Description:
    ------------

    """
    return CompGeoChartJs.ChartJs(self)

  @property
  def plotly(self):
    """
    Description:
    ------------

    """
    return CompGeoPlotly.Plotly(self)

  @property
  def google(self):
    """
    Description:
    ------------

    """
    if not getattr(self.context.rptObj, '_with_google_imports', False):
      raise Exception("Google produce must be added using for example rptObj.imports().google_products(['charts'])")

    return CompGeoGoogle.GeoGoogle(self)



"""
Module dedicated to wrap the Javascript Boolean

Related Pages:

		https://www.w3schools.com/jsref/jsref_valueof_boolean.asp
"""

import json

from epyk.core.js.primitives import JsObject


class JsBoolean(JsObject.JsObject):
  _jsClass = "Boolean"

  def __init__(self, data, varName=None, setVar=False, isPyData=True, report=None):
    if not hasattr(data, 'varName') and isPyData:
      isPyData = True
      data = json.dumps(data)
    super(JsBoolean, self).__init__(data, varName, setVar, isPyData, report)

  @property
  def not_(self):
    """
    Description:
    -----------
    Add the Symbol (!) for the boolean negation.
    This feature is also available directly to the JsObj

    Usage::

      jsObj.objects.boolean.get("weekend").not_

    Related Pages:

			https://developer.mozilla.org/fr/docs/Web/JavaScript/Reference/Op%C3%A9rateurs/Op%C3%A9rateurs_logiques

    :return: The Python Javascript Boolean Object
    """
    self.varName = "!%s" % self.varId
    return self

  def valueOf(self):
    """
    Description:
    -----------
    The valueOf() method returns the primitive value of a boolean.

    Usage::

      jsObj.objects.boolean.new(False, "testBool")
      jsObj.objects.boolean.get("testBool").valueOf()

    Related Pages:

:
    https://www.w3schools.com/jsref/jsref_valueof_boolean.asp

    :return: A Boolean, either "true" or "false"
    """
    return JsBoolean("%s.valueOf()" % self.varId, isPyData=False)




from epyk.core import html

from epyk.interfaces.components import CompLayouts
from epyk.interfaces.components import CompCodes
from epyk.interfaces.components import CompButtons
from epyk.interfaces.components import CompIcons
from epyk.interfaces.components import CompInputs
from epyk.interfaces.components import CompMedia
from epyk.interfaces.components import CompLists
from epyk.interfaces.components import CompNumbers
from epyk.interfaces.tables import CompTables
from epyk.interfaces.graphs import CompCharts
from epyk.interfaces.geo import CompGeo
from epyk.interfaces.components import CompTexts
from epyk.interfaces.components import CompRich
from epyk.interfaces.components import CompImages
from epyk.interfaces.components import CompLinks
from epyk.interfaces.components import CompSliders
from epyk.interfaces.components import CompNetwork
from epyk.interfaces.components import CompForms
from epyk.interfaces.components import CompTags
from epyk.interfaces.components import CompFields
from epyk.interfaces.components import CompTrees
from epyk.interfaces.components import CompVignets
from epyk.interfaces.components import CompMenus
from epyk.interfaces.components import CompPanels
from epyk.interfaces.components import CompModals
from epyk.interfaces.components import CompNavigation
from epyk.interfaces.components import CompSteps
from epyk.interfaces.components import CompDrawers
from epyk.interfaces.components import CompSteppers
from epyk.interfaces.components import CompTitles


class Components(object):
  def __init__(self, rptObj):
    self.rptObj = rptObj

    # Special shortcut for some components
    self.button = self.buttons.button #: shortcut for button :func:`epyk.interfaces.components.CompButtons.Buttons.button`
    self.toggle = self.buttons.toggle #: shortcut for Toogle button :func:`epyk.interfaces.components.CompButtons.Buttons.toggle`
    self.input = self.inputs.input #: shortcut for input :func:`epyk.interfaces.components.CompInputs.Inputs.input`
    self.div = self.layouts.div #: shortcut for div :func:`epyk.interfaces.components.CompLayouts.Layouts.div`
    self.grid = self.layouts.grid #: shortcut for grid :func:`epyk.interfaces.components.CompLayouts.Layouts.grid`
    self.row = self.layouts.row #: shortcut for row :func:`epyk.interfaces.components.CompLayouts.Layouts.row`
    self.col = self.layouts.col #: shortcut for column :func:`epyk.interfaces.components.CompLayouts.Layouts.col`
    self.table = self.tables.tabulator #: shortcut for tabulator :func:`epyk.interfaces.components.CompTables.Tables.tabulator`
    self.pivot = self.tables.pivot #: shortcut for pivot :func:`epyk.interfaces.components.CompTables.Tables.pivot`
    self.text = self.texts.text #: shortcut for text :func:`epyk.interfaces.components.CompTexts.Texts.text`
    self.title = self.texts.title #: shortcut for title :func:`epyk.interfaces.components.CompTexts.Texts.title`
    self.icon = self.images.icon #: shortcut for icon :func:`epyk.interfaces.components.CompImages.Images.icon`
    self.img = self.images.img #: shortcut for img :func:`epyk.interfaces.components.CompImages.Images.img`
    self.list = self.lists.list #: shortcut for list :func:`epyk.interfaces.components.CompLists.Lists.list`
    self.link = self.links.link #: shortcut for link :func:`epyk.interfaces.components.CompLinks.Links.link`
    self.check = self.buttons.check #: shortcut for check :func:`epyk.interfaces.components.CompButtons.Buttons.check`
    self.slider = self.sliders.slider #: shortcut for slider :func:`epyk.interfaces.components.CompSliders.Sliders.slider`
    self.select = self.lists.select #: shortcut for select :func:`epyk.interfaces.components.CompLists.Lists.select`
    self.lookup = self.lists.lookup #: shortcut for lookup :func:`epyk.interfaces.components.CompLists.Lists.lookup`
    self.date = self.fields.cob #: shortcut for date :func:`epyk.interfaces.components.CompFields.Fields.cob`
    self.tree = self.lists.tree #: shortcut for tree :func:`epyk.interfaces.components.CompLists.Lists.tree`
    self.info = self.rich.info #: shortcut for info :func:`epyk.interfaces.components.CompRich.Rich.input`
    self.radio = self.buttons.radio #: shortcut for radio :func:`epyk.interfaces.components.CompButtons.Buttons.radio`
    self.navbar = self.navigation.bar #: shortcut for bar :func:`epyk.interfaces.components.CompNavigation.Navigation.bar`
    self.footer = self.navigation.footer #: shortcut for footer :func:`epyk.interfaces.components.CompNavigation.Navigation.footer`
    self.modal = self.modals.forms #: shortcut for footer :func:`epyk.interfaces.components.CompModals.Modals.forms`
    self.disclaimer = self.modals.disclaimer #: shortcut for footer :func:`epyk.interfaces.components.CompModals.Modals.disclaimer`
    self.drawer = self.drawers.drawer #: shortcut for drawer :func:`epyk.interfaces.components.CompDrawers.Drawers.drawer`
    self.stepper = self.steppers.stepper #: shortcut for steppers :func:`epyk.interfaces.components.CompSteppers.Steppers.stepper`
    self.chips = self.lists.chips #: shortcut for chips :func:`epyk.interfaces.components.CompLists.Lists.chips`
    self.contextual = self.menus.contextual #: shortcut for chips :func:`epyk.interfaces.components.CompMenus.Menus.contextual`
    self.hidden = self.fields.hidden #: shortcut for chips :func:`epyk.interfaces.components.CompInputs.Inputs.input`

    # Shortcut to some important HTML tags
    self.label = self.texts.label
    self.textarea = self.inputs.textarea
    self.header = self.layouts.header
    self.section = self.layouts.section
    self.composite = self.rich.composite

  def css(self, cssAttrs):
    """
    Description:
    ------------
    Change the CSS Style of the main container in the page
    """
    self.rptObj._props.setdefault("css", {})["container"] = cssAttrs

  @property
  def codes(self):
    """
    Description:
    ------------
    Group all the UI Components dedicated to display code fragments

    This will wrap the Javascript module codemirror

    Related Pages:

			https://codemirror.net/doc/manual.html
    """
    return CompCodes.Code(self)

  @property
  def network(self):
    """
    Description:
    ------------
    Group all the UI Components dedicated to display messaging services.

    This category will group (chat, RSS streams, forum, bot ...).
    Those components are interactive and they would require underlying services and databases in order to fully work
    """
    return CompNetwork.Network(self)

  @property
  def sliders(self):
    """
    Description:
    ------------
    Group all the UI components dedicated to produce slider items.

    Those components are interactive and can be used to filter the data on other items in the dashboard.
    Those components are mainly relying on Jquery and JqueryUi
    """
    return CompSliders.Sliders(self)

  @property
  def _3d(self):
    """
    Description:
    ------------
    Group all the 3D charts
    """
    return CompCharts.Chart3d(self)

  @property
  def _2d(self):
    """
    Description:
    ------------
    Group all the 2D charts
    """
    return CompCharts.Chart2d(self)

  @property
  def titles(self):
    """
    Description:
    ------------
    Group all the UI components dedicated to produce titles
    """
    return CompTitles.Titles(self)

  @property
  def links(self):
    """
    Description:
    ------------
    Group all the UI components dedicated to produce links to another page or website.
    """
    return CompLinks.Links(self)

  @property
  def navigation(self):
    """
    Description:
    ------------
    Group all the UI components dedicated to produce navigation components such as navigation bar, footer, banner...
    """
    return CompNavigation.Navigation(self)

  @property
  def banners(self):
    """
    Description:
    ------------
    Group all the available banners
    """
    return CompNavigation.Banners(self)

  @property
  def rich(self):
    """
    Description:
    ------------
    Group all the UI components dedicated to produce rich HTML Components.

    This category will take into account very specific and bespoke components.
    """
    return CompRich.Rich(self)

  @property
  def vignets(self):
    """
    Description:
    ------------
    Group all the UI components dedicated to produce rich HTML Components.

    This category will take into account very specific and bespoke components.
    """
    return CompVignets.Vignets(self)

  @property
  def numbers(self):
    """
    Description:
    ------------
    Group all the UI components dedicated to produce Numbers components.

    The items in this category will not be editable and they will only provide nice number renderings
    """
    return CompNumbers.Numbers(self)

  @property
  def texts(self):
    """
    Description:
    ------------
    Group all the UI components dedicated to produce text components.

    The items in this category will not be editable and they will only provide nice text structure like paragraph,
    formatted text...
    """
    return CompTexts.Texts(self)

  @property
  def images(self):
    """
    Description:
    ------------
    Group all the UI components dedicated to produce image or collection of images.
    """
    return CompImages.Images(self)

  @property
  def lists(self):
    """
    Description:
    ------------
    Group all the UI components dedicated to produce list or selection items.

    Simple list, trees or dropdown boxes will be part of this category of items
    """
    return CompLists.Lists(self)

  @property
  def trees(self):
    """
    Description:
    ------------
    Group all the UI components dedicated to produce Trees or selection items.

    """
    return CompTrees.Trees(self)

  @property
  def geo(self):
    """
    Description:
    ------------
    Group all the UI components dedicated to produce Trees or selection items.

    """
    return CompGeo.Geo(self)

  @property
  def buttons(self):
    """
    Description:
    ------------
    Group all the UI components dedicated to produce button or checkbox.
    """
    return CompButtons.Buttons(self)

  @property
  def tables(self):
    """
    Description:
    ------------
    Group all the UI components dedicated to produce tables or pivot tables.

    Different kind of tables are available in the framework (Tabulator, Datatable, PivotTable or even a bespoke
    implementation).
    """
    return CompTables.Tables(self)

  @property
  def steps(self):
    """
    Description:
    ------------
    Group all the UI steps components.
    """
    return CompSteps.Steppers(self)

  @property
  def drawers(self):
    """
    Description:
    ------------
    Group all the UI drawers components.
    """
    return CompDrawers.Drawers(self)

  @property
  def steppers(self):
    """
    Description:
    ------------
    Group all the UI steppers components.
    """
    return CompSteppers.Steppers(self)

  @property
  def media(self):
    """
    Description:
    ------------
    Group all the UI components dedicated to produce media (video and audio) items.

    Plain Vanilla HTML5 components
    """
    return CompMedia.Media(self)

  @property
  def inputs(self):
    """
    Description:
    ------------
    Group all the UI components dedicated to produce input items.

    Those components are editable items which need to be updated by the user of the dashboard.
    This category will take into account textarea, input text...
    """
    return CompInputs.Inputs(self)

  @property
  def fields(self):
    """
    Description:
    ------------
    Group all the UI components dedicated to produce input items.

    Those components are editable items which need to be updated by the user of the dashboard.
    This category will take into account textarea, input text...
    """
    return CompFields.Fields(self)

  @property
  def timelines(self):
    """
    Description:
    ------------
    """
    return CompFields.Timelines(self)

  @property
  def icons(self):
    """
    Description:
    ------------
    Group all the UI components dedicated to produce icon items.

    This category of component will rely on the font-awesome library for the final display.
    """
    return CompIcons.Icons(self)

  @property
  def menus(self):
    """
    Description:
    ------------
    """
    return CompMenus.Menus(self)

  @property
  def panels(self):
    """
    Description:
    ------------
    """
    return CompPanels.Panels(self)

  @property
  def layouts(self):
    """
    Description:
    ------------
    Group all the UI components dedicated to produce component containers.

    All the items in this category are dedicated for the structure of the dashboard and they
    are mainly holder of other components.
    This will mainly rely on bootstrap for the display of the different objects in the page.
    """
    return CompLayouts.Layouts(self)

  @property
  def forms(self):
    """
    Description:
    ------------
    Group all the Forms components dedicated to drop data.

    Related Pages:

			https://www.w3schools.com/html/html_forms.asp
    """
    return CompForms.Forms(self)

  @property
  def modals(self):
    """
    Description:
    ------------
    Group all the UI components dedicated to produce modal components.
    """
    return CompModals.Modals(self)

  @property
  def charts(self):
    """
    Description:
    ------------
    Group all the UI components dedicated to produce charts.

    Different kind of charts framework are available (ChartJs, Plotly, C3, Billboard, NVD3, DC, Vis or even D3).
    """
    return CompCharts.Graphs(self)

  @property
  def tags(self):
    """
    Description:
    ------------
    Group all the other tags available in HTML.

    Those tags can be considered as normal HTML component, which means Js and CSS features are also available
    """
    return CompTags.Tags(self)

  def contents(self, title="Contents", top=10, right=10, left=None, width=(None, "%"), height=(None, "px"), options=None, profile=None):
    """
    Description:
    ------------
    Add a content table to the page

    Related Pages:

    Attributes:
    ----------
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param profile: Optional. A flag to set the component performance storage
    """
    html_contents = html.HtmlTextComp.ContentsTable(self.rptObj, title, width, height, options, profile)
    html_contents.style.css.top = top
    if left is not None:
      html_contents.style.css.left = left
    else:
      html_contents.style.css.right = right
    # Attach the table content to the main report object
    self.rptObj._content_table = html_contents
    return html_contents

  def bespoke(self, htmlCls, *args, **kwargs):
    """
    Description:
    ------------
    Hook to allow the creation of bespoke component using specific configurations.
    Components can be self contained in a module and rely on external packages

    Tip: Look at the Import.extend function in order to add external Js and CSS modules to your environment

    Attributes:
    ----------
    :param htmlCls: Class. The bespoke HTML component
    :param args: The python attributes used in the HTML component contructor
    :param kwargs: The python attributes used in the HTML component contructor
    """
    return htmlCls(self.rptObj, *args, **kwargs)

  def _tags(self, vals=None, title="", icon="", width=(100, "%"), height=(None, "px"), htmlCode=None, profile=None):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param vals: Optional.
    :param title: Optional.
    :param icon: Optional. A string with the value of the icon to display from font-awesome
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param htmlCode: Optional. An identifier for this component (on both Python and Javascript side)
    :param profile: Optional. A flag to set the component performance storage

    :rtype: html.HtmlTextEditor.Tags
    """
    return html.HtmlTextEditor.Tags(self.rptObj, vals, title, icon, width, height, htmlCode, profile)

  def loading(self, text="Loading", color=None, options=None):
    """
    Description:
    ------------
    Entry point to the loading component

    This component will create a
      - label component for the text
      - icon component for the loading icon

    Attributes:
    ----------
    :param text:
    :param color:
    :param options:
    """
    html_loading = html.HtmlOthers.Loading(self.rptObj, text, color, options or {})
    return html_loading

  def form(self, action=None, method=None, helper=None):
    """
    Description:
    ------------
    Creates an new empty form

    Usage::

      f = rptObj.ui.form()

    Attributes:
    ----------
    :param action:
    :param method:
    :param helper:
    """
    form = html.HtmlContainer.Form(self.rptObj, [], action, method, helper)
    return form

  def json(self, data=None, width=(None, '%'), height=(100, '%'), options=None, profile=None):
    """
    Description:
    ------------
    HTML component to display a Json

    Related Pages:

			https://github.com/mohsen1/json-formatter-js

    Attributes:
    ----------
    :param data: Dictioanry. The Json object to be display
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param options: Optional. Dictionary with the component properties
    :param profile: Boolean
    """
    data = data or {}
    h_json = html.HtmlOthers.HtmlJson(self.rptObj, data, width, height, options, profile)
    if height[1] != '%':
      h_json.style.css.overflow = 'auto'
    return h_json

  def postit(self, components=None, anchor=None):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param components:
    :param anchor: Component. Optional.
    """
    postit = self.rptObj.ui.div()
    if anchor is None:
      anchor = self.rptObj.ui.icon("fas fa-map-marker")
      anchor.style.css.padding = "4px"
      postit += anchor
    postit.anchor = anchor
    popup = self.rptObj.ui.div(components, width=(None, 'px'))
    popup.css({"display": 'none', 'position': 'absolute', 'border': '1px solid black', 'border-radius': '5px',
               'padding': '5px', 'background': self.rptObj.theme.greys[0]})
    postit += popup
    postit.popup = popup
    anchor.mouse([
      popup.dom.position(dx=10, dy=10), popup.dom.css({"display": 'block'}).r
    ],
      [
        popup.dom.css({"display": 'none'}).r]
    )
    return postit

  def extension(self, package_name, alias=None):
    """
    Description:
    ------------
    Add an extension base on it is name

    Attributes:
    ----------
    :param package_name: String. The package name
    :param alias: String. The alias for the link in report.ui
    """
    mod = __import__(package_name)
    __import__("%s.components" % package_name)
    if alias is None:
      alias = getattr(mod.components, 'alias', package_name)
    setattr(self, alias, mod.components.Components(self))
